"""Test asymptotic calculator"""

from spey.hypothesis_testing.asymptotic_calculator import (
    compute_asymptotic_confidence_level,
)
from spey.hypothesis_testing.toy_calculator import compute_toy_confidence_level


def test_compute_asymptotic_confidence_level():
    """Test asymptotic CL computer"""

    a, b = compute_asymptotic_confidence_level(1.0, 2.0, test_stat="qtilde")

    assert a == [0.05933583307142766], "Observed CLs value is wrong."

    assert b == [
        0.05933583307142766,
        0.14339349869880672,
        0.31731050786291404,
        0.5942867086725301,
        0.5942867086725301,
    ], "expected CLs value is wrong."

    a, b = compute_asymptotic_confidence_level(1.0, 2.0, "q0")

    assert a == [0.001349898031630116], "Observed CLs value is wrong."
    assert b == [
        0.001349898031630116,
        0.02275013194817923,
        0.15865525393145702,
        0.5,
        0.5,
    ], "expected CLs value is wrong."


def test_compute_toy_confidence_level():
    """Test toy CLs computer"""
    a, b = compute_toy_confidence_level([2.0], [1.0], 1.5)

    assert a == [0.0], "Observed CLs value is wrong."
    assert b == [1.0] * 5, "expected CLs value is wrong."



import logging
import os
import re
import sys
import textwrap
from typing import Any, Callable, Dict, List, Literal, Optional, Text, Tuple, Union

import numpy as np
import pkg_resources
from semantic_version import SimpleSpec, Version

from spey.base import BackendBase, ConverterBase
from spey.combiner import UnCorrStatisticsCombiner
from spey.interface.statistical_model import StatisticalModel, statistical_model_wrapper
from spey.system import logger
from spey.system.exceptions import PluginError
from spey.system.webutils import ConnectionError, check_updates, get_bibtex

from ._version import __version__
from .about import about
from .utils import ExpectationType

__all__ = [
    "version",
    "StatisticalModel",
    "UnCorrStatisticsCombiner",
    "ExpectationType",
    "AvailableBackends",
    "get_backend",
    "get_backend_metadata",
    "reset_backend_entries",
    "BackendBase",
    "ConverterBase",
    "about",
    "math",
    "check_updates",
    "get_backend_bibtex",
    "cite",
    "set_log_level",
]


def __dir__():
    return __all__


logger.init(LoggerStream=sys.stdout)
log = logging.getLogger("Spey")
log.setLevel(logging.INFO)


def set_log_level(level: Literal[0, 1, 2, 3]) -> None:
    """
    Set log level for spey

    Log level can also be set through terminal by the following command

    .. code::

        export SPEY_LOGLEVEL=3

    value corresponds to the levels shown below.

    Args:
        level (``int``): log level

            * 0: error
            * 1: warning
            * 2: info
            * 3: debug
    """
    log_dict = {
        0: logging.ERROR,
        1: logging.WARNING,
        2: logging.INFO,
        3: logging.DEBUG,
    }
    log.setLevel(log_dict[level])


def version() -> Text:
    """
    Version of ``spey`` package

    Returns:
        ``Text``: version in X.Y.Z format
    """
    return __version__


def _get_backend_entrypoints() -> Dict[Text, pkg_resources.EntryPoint]:
    """Collect plugin entries"""
    return {
        entry.name: entry
        for entry in pkg_resources.iter_entry_points("spey.backend.plugins")
    }


_backend_entries: Dict[Text, pkg_resources.EntryPoint] = _get_backend_entrypoints()
# ! Preinitialise backends, it might be costly to scan the system everytime


def reset_backend_entries() -> None:
    """Scan the system for backends and reset the entries"""
    _backend_entries = _get_backend_entrypoints()


def AvailableBackends() -> List[Text]:  # pylint: disable=C0103
    """
    Returns a list of available backends. The default backends are automatically installed
    with ``spey`` package. To enable other backends, please see the relevant section
    in the documentation.

    .. note::

        This function does not validate the backend. For backend validation please use
        :func:`~spey.get_backend` function.

    Returns:
        ``List[Text]``: list of names of available backends.
    """
    return [*_backend_entries.keys()]


def get_backend(name: Text) -> Callable[[Any], StatisticalModel]:
    """
    Statistical model backend retreiver. Available backend names can be found via
    :func:`~spey.AvailableBackends` function.

    Args:
        name (``Text``): backend identifier. This backend refers to different packages
          that prescribes likelihood function.

    Raises:
        :obj:`~spey.system.exceptions.PluginError`: If the backend is not available
          or the available backend requires different version of ``spey``.
        :obj:`AssertionError`: If the backend does not have necessary metadata.

    Returns:
        ``Callable[[Any, ...], StatisticalModel]``:
        A callable function that takes backend specific arguments and two additional
        keyword arguments ``analysis`` (which is a unique identifier of analysis name as :obj:`str`)
        and ``xsection`` (which is cross section value with a.u.). Details about the function can be
        found in :func:`~spey.statistical_model_wrapper`. This wrapper
        returns a :obj:`~spey.StatisticalModel` object.

    Example:

    .. code-block:: python3
        :linenos:

        >>> import spey; import numpy as np
        >>> stat_wrapper = spey.get_backend("default_pdf.uncorrelated_background")

        >>> data = np.array([1])
        >>> signal = np.array([0.5])
        >>> background = np.array([2.])
        >>> background_unc = np.array([1.1])

        >>> stat_model = stat_wrapper(
        ...     signal_yields=signal,
        ...     background_yields=background,
        ...     data=data,
        ...     covariance_matrix=background_unc,
        ...     analysis="simple_sl",
        ...     xsection=0.123
        ... )
        >>> stat_model.exclusion_confidence_level()

    .. note::

        The documentation of the ``stat_wrapper`` defined above includes the docstring
        of the backend as well. Hence typing ``stat_wrapper?`` in terminal will result with
        complete documentation for the :func:`~spey.statistical_model_wrapper` and
        the backend it self which is in this particular example
        :obj:`~spey.backends.simplifiedlikelihood_backend.interface.SimplifiedLikelihoodInterface`.
    """
    backend = _backend_entries.get(name, False)

    if backend:
        statistical_model = backend.load()

        assert hasattr(
            statistical_model, "spey_requires"
        ), "Backend does not have `'spey_requires'` attribute."

        if getattr(statistical_model, "name", False):
            assert (
                statistical_model.name == name
            ), "The identity of the statistical model is wrongly set."
        else:
            setattr(statistical_model, "name", name)

        if Version(version()) not in SimpleSpec(statistical_model.spey_requires):
            raise PluginError(
                f"The backend {name}, requires spey version {statistical_model.spey_requires}. "
                f"However the current spey version is {__version__}."
            )

        # Initialise converter base models
        if ConverterBase in statistical_model.mro():
            statistical_model = statistical_model()

        return statistical_model_wrapper(statistical_model)

    raise PluginError(
        f"The backend {name} is unavailable. Available backends are "
        + ", ".join(AvailableBackends())
        + "."
    )


def get_backend_metadata(name: Text) -> Dict[Text, Any]:
    """
    Retreive metadata about the backend. This includes citation information,
    doi, author names etc. Available backend names can be found via
    :func:`~spey.AvailableBackends` function.

    Args:
        name (``Text``): backend identifier. This backend refers to different packages
          that prescribes likelihood function.

    Raises:
        ~spey.system.exceptions.PluginError: If the backend does not exists.

    Returns:
        ``Dict[Text, Text]``:
        Metadata about the backend.

    Example:

    .. code-block:: python3

        >>> spey.get_backend_metadata("default_pdf.third_moment_expansion")

    will return the following

    .. code-block:: python3

        >>> {'name': 'default_pdf.third_moment_expansion',
        ... 'author': 'SpeysideHEP',
        ... 'version': '0.0.1',
        ... 'spey_requires': '0.0.1',
        ... 'doi': ['10.1007/JHEP04(2019)064'],
        ... 'arXiv': ['1809.05548']}
    """
    backend = _backend_entries.get(name, False)

    if backend:
        statistical_model = backend.load()
        return {
            "name": getattr(statistical_model, "name", "__unknown_model__"),
            "author": getattr(statistical_model, "author", "__unknown_author__"),
            "version": getattr(statistical_model, "version", "__unknown_version__"),
            "spey_requires": statistical_model.spey_requires,
            "doi": list(getattr(statistical_model, "doi", [])),
            "arXiv": list(getattr(statistical_model, "arXiv", [])),
            "zenodo": list(getattr(statistical_model, "zenodo", [])),
        }

    raise PluginError(
        f"The backend {name} is unavailable. Available backends are "
        + ", ".join(AvailableBackends())
        + "."
    )


def get_backend_bibtex(name: Text) -> Dict[Text, List[Text]]:
    """
    Retreive BibTex entry for backend plug-in if available.

    The bibtext entries are retreived both from Inspire HEP, doi.org and zenodo.
    If the arXiv number matches the DOI the output will include two versions
    of the same reference. If backend does not include an arXiv or DOI number
    it will return an empty list.

    .. versionadded:: 0.1.6

    Args:
        name (``Text``): backend identifier. This backend refers to different packages
          that prescribes likelihood function.

    Returns:
        ``Dict[Text, List[Text]]``:
        BibTex entries for the backend. Keywords include inspire, doi.org and zenodo.

        .. versionchanged:: 0.1.7

            In the previous version, function was returning ``List[Text]`` now it returns
            a ``Dict[Text, List[Text]]`` indicating the source of BibTeX entry.

    """
    # pylint: disable=import-outside-toplevel, W1203
    out = {"inspire": [], "doi.org": [], "zenodo": []}
    meta = get_backend_metadata(name)

    try:
        for arxiv_id in meta.get("arXiv", []):
            tmp = get_bibtex("inspire/arxiv", arxiv_id)
            if tmp != "":
                out["inspire"].append(textwrap.indent(tmp, " " * 4))
            else:
                log.debug(f"Can not find {arxiv_id} in Inspire")
        for doi in meta.get("doi", []):
            tmp = get_bibtex("inspire/doi", doi)
            if tmp == "":
                log.debug(f"Can not find {doi} in Inspire, looking at doi.org")
                tmp = get_bibtex("doi", doi)
                if tmp != "":
                    out["doi.org"].append(tmp)
                else:
                    log.debug(f"Can not find {doi} in doi.org")
            else:
                out["inspire"].append(textwrap.indent(tmp, " " * 4))
        for zenodo_id in meta.get("zenodo", []):
            tmp = get_bibtex("zenodo", zenodo_id)
            if tmp != "":
                out["zenodo"].append(textwrap.indent(tmp, " " * 4))
            else:
                log.debug(f"{zenodo_id} is not a valid zenodo identifier")
    except ConnectionError as err:
        log.error("Can not connect to the internet. Please check your connection.")
        log.debug(str(err))
        return out

    return out


def cite() -> List[Text]:
    """Retreive BibTex information for Spey"""
    try:
        arxiv = textwrap.indent(get_bibtex("inspire/arxiv", "2307.06996"), " " * 4)
        zenodo = get_bibtex("zenodo", "10156353")
        linker = re.search("@software{(.+?),\n", zenodo).group(1)
        zenodo = textwrap.indent(zenodo.replace(linker, "spey_zenodo"), " " * 4)
        return arxiv + "\n\n" + zenodo
    except ConnectionError as err:
        log.error("Can not connect to the internet. Please check your connection.")
        log.debug(str(err))
        return ""


if int(os.environ.get("SPEY_LOGLEVEL", -1)) >= 0:
    set_log_level(int(os.environ.get("SPEY_LOGLEVEL")))

if os.environ.get("SPEY_CHECKUPDATE", "ON").upper() != "OFF":
    check_updates()



"""Test test statistics"""

import numpy as np
import pytest
from spey.hypothesis_testing.test_statistics import (
    get_test_statistic,
    compute_teststatistics,
)
from spey.system.exceptions import UnknownTestStatistics


def test_test_statistic():
    """Validate test statistic functions"""

    logpdf = lambda x: -10.0

    q0 = get_test_statistic("q0")

    assert 0.0 == q0(
        np.random.rand(), -1.0, np.random.rand(), logpdf
    ), "q0 should return zero for negative muhat"
    assert 16.0 == q0(np.random.rand(), 1.0, -2.0, logpdf), "q0 returns wrong value."

    qmu = get_test_statistic("qmu")
    assert 0.0 == qmu(1, 2, -10, logpdf), "qmu should return zero if muhat > mu"
    assert 16.0 == qmu(3, 2, -2, logpdf), "qmu returns wrong value."

    qmu_tilde = get_test_statistic("qmutilde")

    assert 0.0 == qmu_tilde(
        1, 2, -10, logpdf
    ), "qmu_tilde should return zero if muhat > mu"
    assert 16.0 == qmu_tilde(3, 2, -2, logpdf), "qmu_tilde returns wrong value"

    with pytest.raises(
        UnknownTestStatistics, match="Requested test statistics bla does not exist."
    ):
        _ = get_test_statistic("bla")


def test_computation():
    """test compute_teststatistics function"""

    logpdf = lambda x: -10.0

    mu = 3.0
    muhat = 2.0
    muhatA = 2.0
    min_nll = 2.0
    min_nllA = 1.0

    logpdf = lambda x: -10.0
    asimov_logpdf = lambda x: -8

    sqrt_qmu, sqrt_qmuA, delta_teststat = compute_teststatistics(
        mu=mu,
        maximum_likelihood=(muhat, min_nll),
        logpdf=logpdf,
        maximum_asimov_likelihood=(muhatA, min_nllA),
        asimov_logpdf=asimov_logpdf,
        teststat="qtilde",
    )

    assert sqrt_qmu == 4.0, "computation of sqrt(mu) is wrong"
    assert sqrt_qmuA == 3.7416573867739413, "computation of sqrt(muA) is wrong"
    assert delta_teststat == 0.2672612419124244, "computation of delta_teststat is wrong"

    asimov_logpdf = lambda x: -12

    sqrt_qmu, sqrt_qmuA, delta_teststat = compute_teststatistics(
        mu=mu,
        maximum_likelihood=(muhat, min_nll),
        logpdf=logpdf,
        maximum_asimov_likelihood=(muhatA, min_nllA),
        asimov_logpdf=asimov_logpdf,
        teststat="qtilde",
    )

    assert sqrt_qmu == 4.0, "computation of sqrt(mu) is wrong"
    assert sqrt_qmuA == 4.69041575982343, "computation of sqrt(muA) is wrong"
    assert delta_teststat == -0.6904157598234297, "computation of delta_teststat is wrong"



"""Test default pdf plugin"""

import numpy as np
import spey


def test_uncorrelated_background():
    """tester for uncorrelated background model"""

    pdf_wrapper = spey.get_backend("default_pdf.uncorrelated_background")

    data = [36, 33]
    signal_yields = [12.0, 15.0]
    background_yields = [50.0, 48.0]
    background_unc = [12.0, 16.0]

    stat_model = pdf_wrapper(
        signal_yields=signal_yields,
        background_yields=background_yields,
        data=data,
        absolute_uncertainties=background_unc,
        analysis="multi_bin",
        xsection=0.123,
    )

    assert np.isclose(stat_model.poi_upper_limit(), 0.8563345655114185), "POI is wrong."
    assert np.isclose(
        stat_model.exclusion_confidence_level()[0], 0.9701795436411219
    ), "CLs is wrong"


def test_correlated_background():
    """tester for correlated background"""

    pdf_wrapper = spey.get_backend("default_pdf.correlated_background")
    statistical_model = pdf_wrapper(
        signal_yields=[12.0, 15.0],
        background_yields=[50.0, 48.0],
        data=[36, 33],
        covariance_matrix=[[144.0, 13.0], [25.0, 256.0]],
        analysis="example",
        xsection=0.123,
    )

    assert np.isclose(statistical_model.poi_upper_limit(), 0.907104376899138), "POI wrong"
    assert np.isclose(
        statistical_model.exclusion_confidence_level()[0], 0.9635100547173434
    ), "CLs is wrong"
    assert np.isclose(
        statistical_model.sigma_mu(1.0), 0.8456469932844632
    ), "Sigma mu is wrong"


def test_third_moment():
    """tester for the third moment"""

    pdf_wrapper = spey.get_backend("default_pdf.third_moment_expansion")
    statistical_model = pdf_wrapper(
        signal_yields=[12.0, 15.0],
        background_yields=[50.0, 48.0],
        data=[36, 33],
        covariance_matrix=[[144.0, 13.0], [25.0, 256.0]],
        third_moment=[0.5, 0.8],
        analysis="example",
        xsection=0.123,
    )

    assert np.isclose(
        statistical_model.exclusion_confidence_level()[0], 0.9614329616396733
    ), "CLs is wrong"
    assert np.isclose(
        statistical_model.poi_upper_limit(), 0.9221339770245336
    ), "POI is wrong"
    assert np.isclose(
        statistical_model.sigma_mu(1.0), 0.854551194250324
    ), "Sigma mu is wrong"


def test_effective_sigma():
    """tester for the effective sigma"""

    pdf_wrapper = spey.get_backend("default_pdf.effective_sigma")
    statistical_model = pdf_wrapper(
        signal_yields=[12.0, 15.0],
        background_yields=[50.0, 48.0],
        data=[36, 33],
        correlation_matrix=[[1.0, 0.06770833], [0.13020833, 1.0]],
        absolute_uncertainty_envelops=[(10.0, 15.0), (13.0, 18.0)],
        analysis="example",
        xsection=0.123,
    )

    assert np.isclose(
        statistical_model.exclusion_confidence_level()[0], 0.8567802529243093
    ), "CLs is wrong"
    assert np.isclose(
        statistical_model.poi_upper_limit(), 1.5298573610113775
    ), "POI is wrong."
    assert np.isclose(
        statistical_model.sigma_mu(1.0), 1.2152765953701747
    ), "Sigma mu is wrong"


def test_poisson():
    """tester for uncorrelated background model"""

    pdf_wrapper = spey.get_backend("default_pdf.poisson")

    data = [36, 33]
    signal_yields = [12.0, 15.0]
    background_yields = [50.0, 48.0]

    stat_model = pdf_wrapper(
        signal_yields=signal_yields,
        background_yields=background_yields,
        data=data,
        analysis="multi_bin",
        xsection=0.123,
    )

    assert np.isclose(stat_model.poi_upper_limit(), 0.3140867496931846), "POI is wrong."
    assert np.isclose(
        stat_model.exclusion_confidence_level()[0], 0.9999807105228611
    ), "CLs is wrong"
    assert np.isclose(stat_model.sigma_mu(1.0), 0.5573350296644078), "Sigma mu is wrong"


def test_normal():
    """tester for gaussian model"""

    statistical_model = spey.get_backend("default_pdf.normal")(
        signal_yields=[12.0],
        background_yields=[50.0],
        data=[36],
        absolute_uncertainties=[20.0],
    )

    assert np.isclose(
        statistical_model.chi2(poi_test_denominator=0),
        2.0
        * (
            0.5 * ((12.0 + 50.0 - 36.0) ** 2 / 20.0**2)
            - (0.5 * ((50.0 - 36.0) ** 2 / 20.0**2))
        ),
    ), "Gaussian chi2 is wrong"


def test_multivariate_gauss():
    """tester for multivar gauss"""

    signal = np.array([12.0, 15.0])
    bkg = np.array([50.0, 48.0])
    data = np.array([36, 33])
    cov = np.array([[144.0, 13.0], [25.0, 256.0]])

    statistical_model = spey.get_backend("default_pdf.multivariate_normal")(
        signal_yields=signal,
        background_yields=bkg,
        data=data,
        covariance_matrix=cov,
    )

    assert np.isclose(
        statistical_model.chi2(poi_test_denominator=0),
        2.0
        * (
            (0.5 * (signal + bkg - data) @ np.linalg.inv(cov) @ (signal + bkg - data))
            - (0.5 * (bkg - data) @ np.linalg.inv(cov) @ (bkg - data))
        ),
    ), "Multivariate gauss wrong"



import numpy as np
from typing import Tuple, Union

from spey.utils import ExpectationType

__all__ = ["Recorder"]


class Recorder:
    """
    Class for recording likelihood evolution of the statistical model. If the recorder is on,
    repetative values will not be computed and will be automatically extracted from the recorder.
    Recorder can be turned on and off globally using the following;

    .. code-block:: python3

        Recorder.turn_off() # turn recorder off globally
        Recorder.turn_on() # turn recorder on globally

    By default, recorder is off.
    """

    RECORD = False

    def __init__(self):
        # Record style:
        # _poi_test_record
        # Expectation type : { poi test : nll value}
        # _maximum_likelihood_record
        # Expectation type : ( poi test , nll value)
        self._poi_test_record = {
            str(ExpectationType.observed): {},
            str(ExpectationType.aposteriori): {},
            str(ExpectationType.apriori): {},
        }
        self._maximum_likelihood_record = {
            str(ExpectationType.observed): False,
            str(ExpectationType.aposteriori): False,
            str(ExpectationType.apriori): False,
        }
        self._freeze_record = False

    @staticmethod
    def turn_off() -> None:
        """Turn off switch"""
        Recorder.RECORD = False

    @staticmethod
    def turn_on() -> None:
        """Turn on switch"""
        Recorder.RECORD = True

    @staticmethod
    def is_on() -> bool:
        return Recorder.RECORD

    def pause(self):
        self._freeze_record = True
        return self

    def play(self):
        self._freeze_record = False
        return self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.play()

    def get_poi_test(self, expected: ExpectationType, poi_test: float) -> Union[float, bool]:
        """Retrieve NLL value for given poi test and expectation value"""
        poi_test = np.float32(poi_test)
        if self.is_on() and not self._freeze_record:
            nll = self._poi_test_record[str(expected)].get(poi_test, False)
            return nll if nll is not False else False
        else:
            return False

    def get_maximum_likelihood(self, expected: ExpectationType) -> Union[Tuple[float, float], bool]:
        """Retrieve maximum likelihood and fit param"""
        if self.is_on() and not self._freeze_record:
            return self._maximum_likelihood_record[str(expected)]
        else:
            return False

    def record_poi_test(
        self, expected: ExpectationType, poi_test: float, negative_loglikelihood: float
    ) -> None:
        if not self._freeze_record and self.is_on():
            poi_test = np.float32(poi_test)
            negative_loglikelihood = np.float32(negative_loglikelihood)
            self._poi_test_record[str(expected)].update({poi_test: negative_loglikelihood})

    def record_maximum_likelihood(
        self, expected: ExpectationType, poi_test: float, negative_loglikelihood: float
    ) -> None:
        if not self._freeze_record and self.is_on():
            poi_test = np.float32(poi_test)
            negative_loglikelihood = np.float32(negative_loglikelihood)
            self._maximum_likelihood_record[str(expected)] = (poi_test, negative_loglikelihood)



"""
Abstract class for Hypothesis base structure. This class contains necessary
tools to compute exclusion limits and POI upper limits
"""

import logging
from abc import ABC, abstractmethod
from functools import partial
from typing import Any, Callable, Dict, List, Literal, Optional, Text, Tuple, Union

import numpy as np
import tqdm

from spey.hypothesis_testing.asymptotic_calculator import (
    compute_asymptotic_confidence_level,
)
from spey.hypothesis_testing.test_statistics import (
    compute_teststatistics,
    get_test_statistic,
)
from spey.hypothesis_testing.toy_calculator import compute_toy_confidence_level
from spey.hypothesis_testing.upper_limits import find_poi_upper_limit
from spey.system.exceptions import (
    AsimovTestStatZero,
    CalculatorNotAvailable,
    MethodNotAvailable,
)
from spey.utils import ExpectationType

__all__ = ["HypothesisTestingBase"]


def __dir__():
    return __all__


log = logging.getLogger("Spey")

# pylint: disable=W1203,C0103


class HypothesisTestingBase(ABC):
    """
    Abstract class that ensures classes that are performing hypothesis teststing includes certain
    set of function to perform necessary computations. This class gives the ability to compute
    exclusion limits and upper limits for the class inherits it.

    Args:
        ntoys (``int``, default ``1000``): Number of toy samples for hypothesis testing.
          (Only used for toy-based hypothesis testing)
    """

    __slots__ = ["ntoys"]

    def __init__(self, ntoys: int = 1000):
        self.ntoys = ntoys
        """Number of toy samples for sample generator during hypothesis testing"""

    @property
    @abstractmethod
    def is_alive(self) -> bool:
        """Returns True if at least one bin has non-zero signal yield."""
        # This method has to be a property

    @property
    @abstractmethod
    def is_asymptotic_calculator_available(self) -> bool:
        """Check if Asymptotic calculator is available for the backend"""
        # This method has to be a property

    @property
    @abstractmethod
    def is_toy_calculator_available(self) -> bool:
        """Check if Toy calculator is available for the backend"""
        # This method has to be a property

    @property
    @abstractmethod
    def is_chi_square_calculator_available(self) -> bool:
        """Check if chi-square calculator is available for the backend"""
        # This method has to be a property

    @abstractmethod
    def likelihood(
        self,
        poi_test: float = 1.0,
        expected: ExpectationType = ExpectationType.observed,
        return_nll: bool = True,
        data: Optional[Union[List[float], np.ndarray]] = None,
        **kwargs,
    ) -> float:
        r"""
        Compute likelihood of the statistical model

        Args:
            poi_test (:obj:`float`, default :obj:`1.0`): parameter of interest, :math:`\mu`.
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

            return_nll (:obj:`bool`, default :obj:`True`): If ``True`` returns negative log-likelihood,
              else likelihood value.
            data (``Union[List[float], np.ndarray]``, default ``None``): input data that to fit. If
              ``None`` data will be set according to ``expected`` input.
            kwargs: keyword arguments for the optimiser.

        Returns:
            :obj:`float`:
            value of likelihood at fixed :math:`\mu`.
        """

    @abstractmethod
    def maximize_likelihood(
        self,
        return_nll: bool = True,
        expected: ExpectationType = ExpectationType.observed,
        allow_negative_signal: bool = True,
        data: Optional[Union[List[float], np.ndarray]] = None,
        **kwargs,
    ) -> Tuple[float, float]:
        r"""
        Compute maximum of the likelihood.

        Args:
            return_nll (:obj:`bool`, default :obj:`True`): If ``True`` returns negative log-likelihood,
              else likelihood value.
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

            allow_negative_signal (:obj:`bool`, default :obj:`True`): If :obj:`True` :math:`\hat\mu`
              value will be allowed to be negative.
            data (``Union[List[float], np.ndarray]``, default ``None``): input data that to fit. If
              ``None`` data will be set according to ``expected`` input.
            kwargs: keyword arguments for the optimiser.

        Returns:
            :obj:`Tuple[float, float]`:
            value of :math:`\hat\mu` and maximum likelihood.
        """

    @abstractmethod
    def asimov_likelihood(
        self,
        poi_test: float = 1.0,
        expected: ExpectationType = ExpectationType.observed,
        return_nll: bool = True,
        test_statistics: Literal["qtilde", "q", "q0"] = "qtilde",
        **kwargs,
    ) -> float:
        r"""
        Compute likelihood at fixed :math:`\mu` for Asimov data

        Args:
            poi_test (:obj:`float`, default :obj:`1.0`): parameter of interest, :math:`\mu`.
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

            return_nll (`bool`, default `True`): If ``True`` returns negative log-likelihood,
              else likelihood value.
            test_statistics (`Text`, default `"qtilde"`): test statistics.

              * ``'qtilde'``: (default) performs the calculation using the alternative test statistic,
                :math:`\tilde{q}_{\mu}`, see eq. (62) of :xref:`1007.1727`
                (:func:`~spey.hypothesis_testing.test_statistics.qmu_tilde`).

                .. warning::

                    Note that this assumes that :math:`\hat\mu\geq0`, hence :obj:`allow_negative_signal`
                    assumed to be :obj:`False`. If this function has been executed by user, :obj:`spey`
                    assumes that this is taken care of throughout the external code consistently.
                    Whilst computing p-values or upper limit on :math:`\mu` through :obj:`spey` this
                    is taken care of automatically in the backend.

              * ``'q'``: performs the calculation using the test statistic :math:`q_{\mu}`, see
                eq. (54) of :xref:`1007.1727` (:func:`~spey.hypothesis_testing.test_statistics.qmu`).
              * ``'q0'``: performs the calculation using the discovery test statistic, see eq. (47)
                of :xref:`1007.1727` :math:`q_{0}` (:func:`~spey.hypothesis_testing.test_statistics.q0`).

            kwargs: keyword arguments for the optimiser.

        Returns:
            :obj:`float`:
            value of the likelihood.
        """

    @abstractmethod
    def maximize_asimov_likelihood(
        self,
        return_nll: bool = True,
        expected: ExpectationType = ExpectationType.observed,
        test_statistics: Literal["qtilde", "q", "q0"] = "qtilde",
        **kwargs,
    ) -> Tuple[float, float]:
        r"""
        Compute maximum of the likelihood for Asimov data.

        Args:
            return_nll (:obj:`bool`, default :obj:`True`): If ``True`` returns negative log-likelihood,
              else likelihood value.
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

            test_statistics (`Text`, default `"qtilde"`): test statistics.

              * ``'qtilde'``: (default) performs the calculation using the alternative test statistic,
                :math:`\tilde{q}_{\mu}`, see eq. (62) of :xref:`1007.1727`
                (:func:`~spey.hypothesis_testing.test_statistics.qmu_tilde`).

                .. warning::

                    Note that this assumes that :math:`\hat\mu\geq0`, hence `allow_negative_signal`
                    assumed to be `False`. If this function has been executed by user, `spey`
                    assumes that this is taken care of throughout the external code consistently.
                    Whilst computing p-values or upper limit on :math:`\mu` through `spey` this
                    is taken care of automatically in the backend.

              * ``'q'``: performs the calculation using the test statistic :math:`q_{\mu}`, see
                eq. (54) of :xref:`1007.1727` (:func:`~spey.hypothesis_testing.test_statistics.qmu`).
              * ``'q0'``: performs the calculation using the discovery test statistic, see eq. (47)
                of :xref:`1007.1727` :math:`q_{0}` (:func:`~spey.hypothesis_testing.test_statistics.q0`).

            kwargs: keyword arguments for the optimiser.

        Returns:
            :obj:`Tuple[float, float]`:
            value of :math:`\hat\mu` and maximum likelihood.
        """

    def fixed_poi_sampler(
        self,
        poi_test: float,
        size: Optional[int] = None,
        expected: ExpectationType = ExpectationType.observed,
        init_pars: Optional[List[float]] = None,
        par_bounds: Optional[List[Tuple[float, float]]] = None,
        **kwargs,
    ) -> Union[np.ndarray, Callable[[int], np.ndarray]]:
        r"""
        Sample data from the statistical model with fixed parameter of interest.

        Args:
            poi_test (``float``, default ``1.0``): parameter of interest or signal strength,
              :math:`\mu`.
            size (``int``, default ``None``): sample size. If ``None`` a callable function
              will be returned which takes sample size as input.
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.
            init_pars (``List[float]``, default ``None``): initial parameters for the optimiser
            par_bounds (``List[Tuple[float, float]]``, default ``None``): parameter bounds for
              the optimiser.
            kwargs: keyword arguments for the optimiser.

        Raises:
            ~spey.system.exceptions.MethodNotAvailable: If bacend does not have sampler implementation.

        Returns:
            ``Union[np.ndarray, Callable[[int], np.ndarray]]``:
            Sampled data with shape of ``(size, number of bins)`` or callable function to sample from
            directly.
        """
        raise NotImplementedError("This method has not been implemented")

    def chi2(
        self,
        poi_test: float = 1.0,
        poi_test_denominator: Optional[float] = None,
        expected: ExpectationType = ExpectationType.observed,
        allow_negative_signal: bool = False,
        **kwargs,
    ) -> float:
        r"""
        If ``poi_test_denominator=None`` computes

        .. math::

            \chi^2 = -2\log\left(\frac{\mathcal{L}(\mu,\theta_\mu)}{\mathcal{L}(\hat\mu,\hat\theta)}\right)

        else

        .. math::

            \chi^2 = -2\log\left(\frac{\mathcal{L}(\mu,\theta_\mu)}{\mathcal{L}(\mu_{\rm denom},\theta_{\mu_{\rm denom}})}\right)

        where :math:`\mu_{\rm denom}` is ``poi_test_denominator`` which is typically zero to compare signal
        model with the background only model.

        Args:
            poi_test (``float``, default ``1.0``): parameter of interest, :math:`\mu`.
            poi_test_denominator (``float``, default ``None``): parameter of interest for the denominator, :math:`\mu`.
                If ``None`` maximum likelihood will be computed.
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

            allow_negative_signal (``bool``, default ``True``): If ``True`` :math:`\hat\mu`
              value will be allowed to be negative. Only valid when ``poi_test_denominator=None``.
            kwargs: keyword arguments for the optimiser.

        Returns:
            ``float``:
            value of the :math:`\chi^2`.
        """
        if poi_test_denominator is None:
            _, denominator = self.maximize_likelihood(
                expected=expected, allow_negative_signal=allow_negative_signal, **kwargs
            )
        else:
            denominator = self.likelihood(
                poi_test=poi_test_denominator, expected=expected, **kwargs
            )
        log.debug(f"denominator: {denominator}")

        return 2.0 * (
            self.likelihood(poi_test=poi_test, expected=expected, **kwargs) - denominator
        )

    def _prepare_for_hypotest(
        self,
        expected: ExpectationType = ExpectationType.observed,
        test_statistics: Literal["qtilde", "q", "q0"] = "qtilde",
        **kwargs,
    ) -> Tuple[
        Tuple[float, float],
        Callable[[float], float],
        Tuple[float, float],
        Callable[[float], float],
    ]:
        r"""
        Prepare necessary computations for hypothesis testing

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
                p-values to be computed.

                * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
                * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
                * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

            test_statistics (`Text`, default `"qtilde"`): test statistics.

                * ``'qtilde'``: (default) performs the calculation using the alternative test statistic,
                :math:`\tilde{q}_{\mu}`, see eq. (62) of :xref:`1007.1727`
                (:func:`~spey.hypothesis_testing.test_statistics.qmu_tilde`).

                .. warning::

                    Note that this assumes that :math:`\hat\mu\geq0`, hence `allow_negative_signal`
                    assumed to be `False`. If this function has been executed by user, :obj:`spey`
                    assumes that this is taken care of throughout the external code consistently.
                    Whilst computing p-values or upper limit on :math:`\mu` through `spey` this
                    is taken care of automatically in the backend.

                * ``'q'``: performs the calculation using the test statistic :math:`q_{\mu}`, see
                eq. (54) of :xref:`1007.1727` (:func:`~spey.hypothesis_testing.test_statistics.qmu`).
                * ``'q0'``: performs the calculation using the discovery test statistic, see eq. (47)
                of :xref:`1007.1727` :math:`q_{0}` (:func:`~spey.hypothesis_testing.test_statistics.q0`).

            kwargs: keyword arguments for the optimiser.

              * **init_pars** (``List[float]``, default ``None``): initial parameters for the optimiser
              * **par_bounds** (``List[Tuple[float, float]]``, default ``None``): parameter bounds for
                the optimiser.

        Returns:
            :obj:`Tuple[ Tuple[float, float], Callable[[float], float], Tuple[float, float], Callable[[float], float]]`:
            (:math:`\hat\mu`, :math:`\arg\min(-\log\mathcal{L})`), :math:`\log\mathcal{L(\mu, \theta_\mu)}`,
            (:math:`\hat\mu_A`, :math:`\arg\min(-\log\mathcal{L}_A)`), :math:`\log\mathcal{L_A(\mu, \theta_\mu)}`
        """
        allow_negative_signal = test_statistics in ["q" or "qmu"]

        muhat, nll = self.maximize_likelihood(
            expected=expected,
            allow_negative_signal=allow_negative_signal,
            **kwargs,
        )
        log.debug(f"muhat: {muhat}, nll: {nll}")
        muhatA, nllA = self.maximize_asimov_likelihood(
            expected=expected,
            test_statistics=test_statistics,
            **kwargs,
        )
        log.debug(f"muhatA: {muhatA}, nllA: {nllA}")

        def logpdf(mu: Union[float, np.ndarray]) -> float:
            return -self.likelihood(
                poi_test=float(mu) if isinstance(mu, (float, int)) else mu[0],
                expected=expected,
                **kwargs,
            )

        def logpdf_asimov(mu: Union[float, np.ndarray]) -> float:
            return -self.asimov_likelihood(
                poi_test=float(mu) if isinstance(mu, (float, int)) else mu[0],
                expected=expected,
                test_statistics=test_statistics,
                **kwargs,
            )

        return (muhat, nll), logpdf, (muhatA, nllA), logpdf_asimov

    def sigma_mu(
        self,
        poi_test: float,
        expected: ExpectationType = ExpectationType.observed,
        test_statistics: Literal["qtilde", "q", "q0"] = "qtilde",
        **kwargs,
    ) -> float:
        r"""
        If available, :math:`\sigma_\mu` will be computed through Hessian of negative log-likelihood
        see :func:`spey.StatisticalModel.sigma_mu_from_hessian` for details.
        However, if not available it will be estimated via :math:`q_{\mu,A}`

        .. math::

            \sigma^2_A = \frac{(\mu - \mu^\prime)^2}{q_{\mu,A}}\quad , \quad q_{\mu,A} = -2\ln\lambda_A(\mu)

        see eq. (31) in :xref:`1007.1727`

        Args:
            poi_test (:obj:`float`, default :obj:`1.0`): parameter of interest, :math:`\mu`.
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

            test_statistics (`Text`, default `"qtilde"`): test statistics.

              * ``'qtilde'``: (default) performs the calculation using the alternative test statistic,
                :math:`\tilde{q}_{\mu}`, see eq. (62) of :xref:`1007.1727`
                (:func:`~spey.hypothesis_testing.test_statistics.qmu_tilde`).

                .. warning::

                    Note that this assumes that :math:`\hat\mu\geq0`, hence `allow_negative_signal`
                    assumed to be `False`. If this function has been executed by user, `spey`
                    assumes that this is taken care of throughout the external code consistently.
                    Whilst computing p-values or upper limit on :math:`\mu` through `spey` this
                    is taken care of automatically in the backend.

              * ``'q'``: performs the calculation using the test statistic :math:`q_{\mu}`, see
                eq. (54) of :xref:`1007.1727` (:func:`~spey.hypothesis_testing.test_statistics.qmu`).
              * ``'q0'``: performs the calculation using the discovery test statistic, see eq. (47)
                of :xref:`1007.1727` :math:`q_{0}` (:func:`~spey.hypothesis_testing.test_statistics.q0`).

            kwargs: keyword arguments for the optimiser.

              * **init_pars** (``List[float]``, default ``None``): initial parameters for the optimiser
              * **par_bounds** (``List[Tuple[float, float]]``, default ``None``): parameter bounds for
                the optimiser.

        Returns:
            :obj:`float`:
            value of the variance on :math:`\mu`.
        """
        if hasattr(self, "sigma_mu_from_hessian"):
            try:
                return self.sigma_mu_from_hessian(
                    poi_test=poi_test, expected=expected, **kwargs
                )
            except MethodNotAvailable:
                log.warning(
                    "Hessian implementation is not available for this backend, "
                    "continuing with the approximate method."
                )
        teststat_func = get_test_statistic(test_statistics)

        muhatA, min_nllA = self.maximize_asimov_likelihood(
            expected=expected, test_statistics=test_statistics, **kwargs
        )
        log.debug(f"muhatA: {muhatA}, min_nllA: {min_nllA}")

        def logpdf_asimov(mu: Union[float, np.ndarray]) -> float:
            return -self.asimov_likelihood(
                poi_test=mu if isinstance(mu, float) else mu[0],
                expected=expected,
                test_statistics=test_statistics,
                **kwargs,
            )

        qmuA = teststat_func(poi_test, muhatA, -min_nllA, logpdf_asimov)

        return 1.0 if qmuA <= 0.0 else np.true_divide(poi_test, np.sqrt(qmuA))

    def exclusion_confidence_level(
        self,
        poi_test: float = 1.0,
        expected: ExpectationType = ExpectationType.observed,
        allow_negative_signal: bool = False,
        calculator: Literal["asymptotic", "toy", "chi_square"] = "asymptotic",
        **kwargs,
    ) -> List[float]:
        r"""
        Compute exclusion confidence level (:math:`1-CL_s`) at a given POI, :math:`\mu`.

        Args:
            poi_test (``float``, default ``1.0``): parameter of interest, :math:`\mu`.
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed. If ``expected`` is set to ``"all"`` it will return both p-values
              and expected p-values where the likelihood will be fitted to observations.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).

                .. note::

                    In case of :obj:`~spey.ExpectationType.observed`, function will return one value
                    which has been fit to the observed data.


              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.

                .. note::

                    In case of :obj:`~spey.ExpectationType.aposteriori`, function will return five value
                    for expected p-values which has been fit to the observed data. Values represent
                    :math:`1\sigma` and :math:`2\sigma` fluctuations from the background. The order of the
                    output order is :math:`-2\sigma` value, :math:`-1\sigma` value, central value,
                    :math:`1\sigma` and :math:`2\sigma` value.

              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

                .. note::

                    In case of :obj:`~spey.ExpectationType.apriori`, function will return five value
                    for expected p-values which has been fit to the SM background. Values represent
                    :math:`1\sigma` and :math:`2\sigma` fluctuations from the background. The
                    output order is :math:`-2\sigma` value, :math:`-1\sigma` value, central value,
                    :math:`1\sigma` and :math:`2\sigma` value.

            allow_negative_signal (``bool``, default ``False``): If ``True`` :math:`\hat\mu`
              value will be allowed to be negative.
            calculator (``Literal["asymptotic", "toy", "chi_square"]``, default ``"asymptotic"``):
              Chooses the computation basis for hypothesis testing

              * ``"asymptotic"``: Uses asymptotic hypothesis testing to compute p-values.
              * ``"toy"``: Uses generated toy samples to compute p-values.
              * ``"chi_square"``: Computes p-values via chi-square;
                :math:`\chi^2=-2\log\frac{\mathcal{L}(1,\theta_1)}{\mathcal{L}(0,\theta_0)}`.

            kwargs: keyword arguments for the optimiser.

              * **init_pars** (``List[float]``, default ``None``): initial parameters for the optimiser
              * **par_bounds** (``List[Tuple[float, float]]``, default ``None``): parameter bounds for
                the optimiser.

        Raises:
          :obj:`~spey.system.exceptions.CalculatorNotAvailable`: If calculator is not available.

        Returns:
            ``List[float]``:
            Exclusion confidence level i.e. :math:`1-CL_s`.
        """
        if not getattr(self, f"is_{calculator}_calculator_available", False):
            raise CalculatorNotAvailable(f"{calculator} calculator is not available.")

        test_stat = "q" if allow_negative_signal else "qtilde"
        verbose = kwargs.pop("verbose", True)

        # NOTE Improve code efficiency, these are not necessary for asymptotic calculator
        if calculator in ["toy", "chi_square"]:
            test_stat_func = get_test_statistic(test_stat)

            def logpdf(
                mu: Union[float, np.ndarray], data: Union[float, np.ndarray]
            ) -> float:
                """Compute logpdf with respect to poi and given data"""
                return -self.likelihood(
                    poi_test=float(mu) if isinstance(mu, (float, int)) else mu[0],
                    expected=expected,
                    data=data,
                    **kwargs,
                )

            def maximize_likelihood(
                data: Union[float, np.ndarray]
            ) -> Tuple[float, float]:
                """Compute maximum likelihood with respect to given data"""
                return self.maximize_likelihood(
                    expected=expected,
                    allow_negative_signal=allow_negative_signal,
                    data=data,
                    **kwargs,
                )

            muhat, min_negloglike = maximize_likelihood(None)

        if calculator == "asymptotic":
            (
                maximum_likelihood,
                logpdf,
                maximum_asimov_likelihood,
                logpdf_asimov,
            ) = self._prepare_for_hypotest(
                expected=expected,
                test_statistics=test_stat,
                **kwargs,
            )

            try:
                _, sqrt_qmuA, delta_teststat = compute_teststatistics(
                    poi_test,
                    maximum_likelihood,
                    logpdf,
                    maximum_asimov_likelihood,
                    logpdf_asimov,
                    test_stat,
                )
                log.debug(
                    f"<asymptotic> sqrt_qmuA = {sqrt_qmuA}, test statistic = {delta_teststat}"
                )

                pvalues, expected_pvalues = compute_asymptotic_confidence_level(
                    sqrt_qmuA, delta_teststat, test_stat
                )
                log.debug(f"pval = {pvalues}, expected pval = {expected_pvalues}")
            except AsimovTestStatZero as err:
                log.error(str(err))
                pvalues, expected_pvalues = [1.0], [1.0] * 5

        elif calculator == "toy":
            signal_samples = self.fixed_poi_sampler(
                poi_test=poi_test, size=self.ntoys, expected=expected, **kwargs
            )

            bkg_samples = self.fixed_poi_sampler(
                poi_test=0.0, size=self.ntoys, expected=expected, **kwargs
            )

            signal_like_test_stat, bkg_like_test_stat = [], []
            with tqdm.tqdm(
                total=self.ntoys,
                unit="toy sample",
                bar_format="{l_bar}{bar:20}{r_bar}{bar:-20b}",
                disable=not verbose,
            ) as pbar:
                for sig_smp, bkg_smp in zip(signal_samples, bkg_samples):
                    muhat_s_b, min_negloglike_s_b = maximize_likelihood(data=sig_smp)
                    signal_like_test_stat.append(
                        test_stat_func(
                            poi_test,
                            muhat_s_b,
                            -min_negloglike_s_b,
                            partial(logpdf, data=sig_smp),
                        )
                    )

                    muhat_b, min_negloglike_b = maximize_likelihood(data=bkg_smp)
                    bkg_like_test_stat.append(
                        test_stat_func(
                            poi_test,
                            muhat_b,
                            -min_negloglike_b,
                            partial(logpdf, data=bkg_smp),
                        )
                    )

                    pbar.update()

            pvalues, expected_pvalues = compute_toy_confidence_level(
                signal_like_test_stat,
                bkg_like_test_stat,
                test_statistic=test_stat_func(
                    poi_test, muhat, -min_negloglike, partial(logpdf, data=None)
                ),
                test_stat=test_stat,
            )
            pvalues, expected_pvalues = np.clip(pvalues, 0.0, 1.0), np.clip(
                expected_pvalues, 0.0, 1.0
            )

        elif calculator == "chi_square":
            ts_s_b = test_stat_func(
                poi_test, muhat, -min_negloglike, partial(logpdf, data=None)
            )
            null_logpdf = logpdf(0.0, None)
            max_logpdf = (
                -min_negloglike if muhat >= 0.0 or test_stat == "q" else null_logpdf
            )
            ts_b_only = np.clip(-2.0 * (null_logpdf - max_logpdf), 0.0, None)
            log.debug(
                f"<chi_square> test statistic: null hypothesis={ts_b_only}, s+b={ts_s_b}"
            )

            delta_ts = None
            sqrt_ts_s_b, sqrt_ts_b_only = np.sqrt(ts_s_b), np.sqrt(ts_b_only)
            if test_stat == "q" or sqrt_ts_s_b <= sqrt_ts_b_only:
                delta_ts = sqrt_ts_b_only - sqrt_ts_s_b
            else:
                try:
                    delta_ts = (ts_b_only - ts_s_b) / (2.0 * sqrt_ts_s_b)
                except ZeroDivisionError:
                    log.error(
                        "Lack of evidence for a signal or deviation from a null hypothesis."
                    )
            if delta_ts is not None:
                pvalues, expected_pvalues = compute_asymptotic_confidence_level(
                    sqrt_ts_s_b, delta_ts, test_stat=test_stat
                )
            else:
                pvalues, expected_pvalues = [1.0], [1.0] * 5

        if expected == "all":
            return list(map(lambda x: 1.0 - x, pvalues)), list(
                map(lambda x: 1.0 - x, expected_pvalues)
            )

        return list(
            map(
                lambda x: 1.0 - x,
                pvalues if expected == ExpectationType.observed else expected_pvalues,
            )
        )

    def significance(
        self, expected: ExpectationType = ExpectationType.observed, **kwargs
    ) -> Tuple[float, float, List[float], List[float]]:
        r"""
        Compute the discovery of a positive signal. See :xref:`1007.1727` eq. (53).
        and sec. 5.1.

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

                .. note::

                    Since :obj:`~spey.ExpectationType.aposteriori` and :obj:`~spey.ExpectationType.observed`
                    are both represent post-fit computation the result will be the same. The only difference
                    can be seen via prefit, :obj:`~spey.ExpectationType.apriori`, computation.

            kwargs: keyword arguments for the optimiser.

              * **init_pars** (``List[float]``, default ``None``): initial parameters for the optimiser
              * **par_bounds** (``List[Tuple[float, float]]``, default ``None``): parameter bounds for
                the optimiser.

        Returns:
            ``Tuple[float, float, List[float], List[float]]``:
            (:math:`\sqrt{q_{0,A}}`, :math:`\sqrt{q_0}`, p-values and expected p-values)
        """
        (
            maximum_likelihood,
            logpdf,
            maximum_asimov_likelihood,
            logpdf_asimov,
        ) = self._prepare_for_hypotest(expected=expected, test_statistics="q0", **kwargs)

        sqrt_q0, sqrt_q0A, delta_teststat = compute_teststatistics(
            0.0,
            maximum_likelihood,
            logpdf,
            maximum_asimov_likelihood,
            logpdf_asimov,
            "q0",
        )
        pvalues, expected_pvalues = compute_asymptotic_confidence_level(
            sqrt_q0A, delta_teststat, "q0"
        )

        return sqrt_q0A, sqrt_q0, pvalues, expected_pvalues

    def poi_upper_limit(
        self,
        expected: ExpectationType = ExpectationType.observed,
        confidence_level: float = 0.95,
        low_init: Optional[float] = 1.0,
        hig_init: Optional[float] = 1.0,
        expected_pvalue: Literal["nominal", "1sigma", "2sigma"] = "nominal",
        maxiter: int = 10000,
        optimiser_arguments: Optional[Dict[Text, Any]] = None,
    ) -> Union[float, List[float]]:
        r"""
        Compute the upper limit for the parameter of interest i.e. :math:`\mu`.

        .. note::

            This function uses ``"qtilde"`` test statistic which means signal values are always
            assumed to be positive i.e. :math:`\hat\mu>0`.

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

            confidence_level (``float``, default ``0.95``): Determines the confidence level of the upper
              limit i.e. the value of :math:`1-CL_s`. It needs to be between ``[0,1]``.
            low_init (``Optional[float]``, default ``1.0``): Lower limit for the search algorithm to start
              If ``None`` it the lower limit will be determined by :math:`\hat\mu + 1.5\sigma_{\hat\mu}`.

              .. note::

                :math:`\sigma_{\hat\mu}` is determined via
                :func:`~spey.base.hypotest_base.HypothesisTestingBase.sigma_mu` function.

            hig_init (``Optional[float]``, default ``1.0``): Upper limit for the search algorithm to start
              If ``None`` it the upper limit will be determined by :math:`\hat\mu + 2.5\sigma_{\hat\mu}`.

              .. note::

                :math:`\sigma_{\hat\mu}` is determined via
                :func:`~spey.base.hypotest_base.HypothesisTestingBase.sigma_mu` function.

            expected_pvalue (``Literal["nominal", "1sigma", "2sigma"]``, default ``"nominal"``):
              In case of :obj:`~spey.ExpectationType.aposteriori` and :obj:`~spey.ExpectationType.apriori`
              expectation, gives the choice to find excluded upper limit for statistical deviations as well.

              * ``"nominal"``: only find the upper limit for the central p-value. Returns a single value.
              * ``"1sigma"``: find the upper limit for central p-value and :math:`1\sigma` fluctuation from
                background. Returns 3 values.
              * ``"2sigma"``: find the upper limit for central p-value and :math:`1\sigma` and
                :math:`2\sigma` fluctuation from background. Returns 5 values.

              .. note::

                For ``expected=spey.ExpectationType.observed``, ``expected_pvalue`` argument will
                be overwritten to ``"nominal"``.

            maxiter (``int``, default ``10000``): Maximum iteration limit for the optimiser.
            optimiser_arguments (``Dict``, default ``None``): Arguments for optimiser that is used
              to compute likelihood and its maximum.

        Returns:
            ``Union[float, List[float]]``:
            In case of nominal values it returns a single value for the upper limit. In case of
            ``expected_pvalue="1sigma"`` or ``expected_pvalue="2sigma"`` it will return a list of
            multiple upper limit values for fluctuations as well as the central value. The
            output order is :math:`-2\sigma` value, :math:`-1\sigma` value, central value,
            :math:`1\sigma` and :math:`2\sigma` value.
        """
        assert (
            0.0 <= confidence_level <= 1.0
        ), "Confidence level must be between zero and one."
        expected_pvalue = (
            "nominal" if expected == ExpectationType.observed else expected_pvalue
        )

        # If the signal yields in all regions are zero then return inf.
        # This means we are not able to set a bound with the given information.
        if not self.is_alive:
            return {"nominal": np.inf, "1sigma": [np.inf] * 3, "2sigma": [np.inf] * 5}[
                expected_pvalue
            ]

        optimiser_arguments = optimiser_arguments or {}
        (
            maximum_likelihood,
            logpdf,
            maximum_asimov_likelihood,
            logpdf_asimov,
        ) = self._prepare_for_hypotest(
            expected=expected,
            test_statistics="qtilde",
            **optimiser_arguments,
        )

        if None in [low_init, hig_init]:
            muhat = maximum_likelihood[0] if maximum_likelihood[0] > 0.0 else 0.0
            sigma_mu = (
                self.sigma_mu(muhat, expected=expected)
                if not np.isclose(muhat, 0.0)
                else 1.0
            )
            low_init = low_init or muhat + 1.5 * sigma_mu
            hig_init = hig_init or muhat + 2.5 * sigma_mu
            log.debug(f"new low_init = {low_init}, new hig_init = {hig_init}")

        return find_poi_upper_limit(
            maximum_likelihood=maximum_likelihood,
            logpdf=logpdf,
            maximum_asimov_likelihood=maximum_asimov_likelihood,
            asimov_logpdf=logpdf_asimov,
            expected=expected,
            confidence_level=confidence_level,
            allow_negative_signal=False,
            low_init=low_init,
            hig_init=hig_init,
            expected_pvalue=expected_pvalue,
            maxiter=maxiter,
        )



"""Tools for computing confidence level and pvalues at the asymptotic limit"""

from typing import List, Text, Tuple

from .distributions import AsymptoticTestStatisticsDistribution
from .utils import expected_pvalues, pvalues

__all__ = ["compute_asymptotic_confidence_level"]


def __dir__():
    return __all__


def compute_asymptotic_confidence_level(
    sqrt_qmuA: float, delta_test_statistic: float, test_stat: Text = "qtilde"
) -> Tuple[List[float], List[float]]:
    r"""
    Compute p values i.e. :math:`p_{s+b}`, :math:`p_b` and :math:`p_s`
        
    .. math::
    
        p_{s+b}&=& \int_{-\infty}^{-\sqrt{q_{\mu,A}} - \Delta q_\mu} \mathcal{N}(x| 0, 1) dx \\
        p_{b}&=& \int_{-\infty}^{-\Delta q_\mu} \mathcal{N}(x| 0, 1) dx \\
        p_{s} &=& p_{s+b}/ p_{b}

    where :math:`q_\mu` stands for the test statistic and A stands for Assimov.
    
    .. math::
        
        \Delta q_\mu = \begin{cases} 
        \sqrt{q_{\mu}} - \sqrt{q_{\mu,A}}, &  \text{if}\ \sqrt{q_{\mu}} \leq \sqrt{q_{\mu,A}} \\
        \frac{\sqrt{q_{\mu}} - \sqrt{q_{\mu,A}}}{2\ \sqrt{q_{\mu,A}}}, & \text{otherwise} 
        \end{cases}

    Note that the CDF has a cutoff at :math:`-\sqrt{q_{\mu,A}}`, hence if 
    :math:`p_{s\ {\rm or}\ s+b} < -\sqrt{q_{\mu,A}}` p-value will not be computed.

    .. seealso::
    
        eq. 66 of :xref:`1007.1727`

    Args:
        sqrt_qmuA (``float``): test statistic for Asimov data :math:`\sqrt{q_{\mu,A}}`.
        delta_test_statistic (``float``): :math:`\Delta q_\mu`
        test_stat (``Text``, default ``"qtilde"``): test statistics.

          * ``'qtilde'``: (default) performs the calculation using the alternative test statistic,
            :math:`\tilde{q}_{\mu}`, see eq. (62) of :xref:`1007.1727`
            (:func:`~spey.hypothesis_testing.test_statistics.qmu_tilde`).

            .. warning::

                Note that this assumes that :math:`\hat\mu\geq0`, hence :obj:`allow_negative_signal`
                assumed to be ``False``. If this function has been executed by user, :obj:`spey`
                assumes that this is taken care of throughout the external code consistently.
                Whilst computing p-values or upper limit on :math:`\mu` through :obj:`spey` this
                is taken care of automatically in the backend.

          * ``'q'``: performs the calculation using the test statistic :math:`q_{\mu}`, see
            eq. (54) of :xref:`1007.1727` (:func:`~spey.hypothesis_testing.test_statistics.qmu`).
          * ``'q0'``: performs the calculation using the discovery test statistic, see eq. (47)
            of :xref:`1007.1727` :math:`q_{0}` (:func:`~spey.hypothesis_testing.test_statistics.q0`).

    Returns:
        ``Tuple[List[float], List[float]]``:
        returns p-values and expected p-values.
        
    .. seealso::

        :func:`~spey.hypothesis_testing.test_statistics.compute_teststatistics`, 
        :func:`~spey.hypothesis_testing.utils.pvalues`,
        :func:`~spey.hypothesis_testing.utils.expected_pvalues`
    """
    cutoff = -sqrt_qmuA  # use clipped normal -> normal will mean -np.inf
    # gives more stable result for cases that \hat\mu > \mu : see eq 14, 16 :xref:`1007.1727`

    sig_plus_bkg_distribution = AsymptoticTestStatisticsDistribution(-sqrt_qmuA, cutoff)
    bkg_only_distribution = AsymptoticTestStatisticsDistribution(0.0, cutoff)

    CLsb_obs, CLb_obs, CLs_obs = pvalues(
        delta_test_statistic, sig_plus_bkg_distribution, bkg_only_distribution
    )
    CLsb_exp, CLb_exp, CLs_exp = expected_pvalues(
        sig_plus_bkg_distribution, bkg_only_distribution
    )

    return ([CLsb_obs], CLsb_exp) if test_stat == "q0" else ([CLs_obs], CLs_exp)



from typing import List, Text, Tuple

import numpy as np

from .distributions import EmpricTestStatisticsDistribution
from .utils import expected_pvalues, pvalues

__all__ = ["compute_toy_confidence_level"]


def __dir__():
    return __all__


def compute_toy_confidence_level(
    signal_like_test_statistic: List[float],
    background_like_test_statistic: List[float],
    test_statistic: float,
    test_stat: Text = "qtilde",
) -> Tuple[List[float], List[float]]:
    r"""
    Compute confidence limits i.e. :math:`CL_{s+b}`, :math:`CL_b` and :math:`CL_s`

    Args:
        signal_like_test_statistic (``List[float]``): signal like test statistic values
        background_like_test_statistic (``List[float]``): background like test statistic values
        test_statistic (``float``): value for parameter of interest
        test_stat (``Text``, default ``"qtilde"``): test statistics.

          * ``'qtilde'``: (default) performs the calculation using the alternative test statistic,
            :math:`\tilde{q}_{\mu}`, see eq. (62) of :xref:`1007.1727`
            (:func:`~spey.hypothesis_testing.test_statistics.qmu_tilde`).

            .. warning::

                Note that this assumes that :math:`\hat\mu\geq0`, hence :obj:`allow_negative_signal`
                assumed to be ``False``. If this function has been executed by user, :obj:`spey`
                assumes that this is taken care of throughout the external code consistently.
                Whilst computing p-values or upper limit on :math:`\mu` through :obj:`spey` this
                is taken care of automatically in the backend.

          * ``'q'``: performs the calculation using the test statistic :math:`q_{\mu}`, see
            eq. (54) of :xref:`1007.1727` (:func:`~spey.hypothesis_testing.test_statistics.qmu`).
          * ``'q0'``: performs the calculation using the discovery test statistic, see eq. (47)
            of :xref:`1007.1727` :math:`q_{0}` (:func:`~spey.hypothesis_testing.test_statistics.q0`).


    Returns:
        ``Tuple[List[float], List[float]]``:
        returns p-values and expected p-values.
    """

    sig_plus_bkg_distribution = EmpricTestStatisticsDistribution(
        np.array(signal_like_test_statistic, dtype=np.float64)
    )
    bkg_only_distribution = EmpricTestStatisticsDistribution(
        np.array(background_like_test_statistic, dtype=np.float64)
    )

    CLsb_obs, CLb_obs, CLs_obs = pvalues(
        test_statistic, sig_plus_bkg_distribution, bkg_only_distribution
    )
    CLsb_exp, CLb_exp, CLs_exp = expected_pvalues(
        sig_plus_bkg_distribution, bkg_only_distribution
    )

    return ([CLsb_obs], CLsb_exp) if test_stat == "q0" else ([CLs_obs], CLs_exp)



"""Logging style for spey"""
import logging
import sys

# pylint: disable=C0103


class ColoredFormatter(logging.Formatter):
    """Coloured logging formatter for spey"""

    def __init__(self, msg):
        logging.Formatter.__init__(self, msg)

    def format(self, record):
        if record.levelno >= 50:  # FATAL
            color = "\x1b[31mSpey - ERROR: "
        elif record.levelno >= 40:  # ERROR
            color = "\x1b[31mSpey - ERROR: "
        elif record.levelno >= 30:  # WARNING
            color = "\x1b[35mSpey - WARNING: "
        elif record.levelno >= 20:  # INFO
            color = "\x1b[0mSpey: "
        elif record.levelno >= 10:  # DEBUG
            color = (
                f"\x1b[36mSpey - DEBUG ({record.module}.{record.funcName}() "
                f"in {record.filename}::L{record.lineno}): "
            )
        else:  # ANYTHING ELSE
            color = "\x1b[0mSpey: "

        record.msg = color + str(record.msg) + "\x1b[0m"
        return logging.Formatter.format(self, record)


def init(LoggerStream=sys.stdout):
    """Initialise logger"""
    rootLogger = logging.getLogger()
    hdlr = logging.StreamHandler()
    fmt = ColoredFormatter("%(message)s")
    hdlr.setFormatter(fmt)
    rootLogger.addHandler(hdlr)

    # we need to replace all root loggers by ma5 loggers for a proper
    # interface with madgraph5
    SpeyLogger = logging.getLogger("Spey")
    for hdlr in SpeyLogger.handlers:
        SpeyLogger.removeHandler(hdlr)
    hdlr = logging.StreamHandler(LoggerStream)
    fmt = ColoredFormatter("%(message)s")
    hdlr.setFormatter(fmt)
    SpeyLogger.addHandler(hdlr)
    SpeyLogger.propagate = False



"""Various helper functions"""

import numpy as np


def correlation_to_covariance(
    correlation_matrix: np.ndarray, standard_deviations: np.ndarray
) -> np.ndarray:
    """
    Convert correlation matrix into covariance matrix

    Args:
        correlation_matrix (``np.ndarray``): a real NxN matrix
        standard_deviations (``np.ndarray``): a real N-dimensional
          vector representing standard deviations.

    Returns:
        ``np.ndarray``:
        Covariance matrix
    """
    sigma = np.diag(standard_deviations)
    return sigma @ correlation_matrix @ sigma


def covariance_to_correlation(covariance_matrix: np.ndarray) -> np.ndarray:
    """
    Convert covariance matrix into correlation matrix.

    Args:
        covariance_matrix (``np.ndarray``): a real NxN matrix

    Returns:
        ``np.ndarray``:
        Correlation matrix
    """
    sigma = np.linalg.inv(np.diag(np.sqrt(np.diag(covariance_matrix))))
    return sigma @ covariance_matrix @ sigma



import logging
from typing import Callable, Dict, List, Optional, Tuple

import numpy as np

from spey.base.model_config import ModelConfig

from .scipy_tools import minimize

# pylint: disable=W1203

log = logging.getLogger("Spey")


def fit(
    func: Callable[[np.ndarray], np.ndarray],
    model_configuration: ModelConfig,
    do_grad: bool = False,
    hessian: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    initial_parameters: Optional[np.ndarray] = None,
    bounds: Optional[List[Tuple[float, float]]] = None,
    fixed_poi_value: Optional[float] = None,
    logpdf: Optional[Callable[[List[float]], float]] = None,
    constraints: Optional[List[Dict]] = None,
    **options,
) -> Tuple[float, np.ndarray]:
    """
    Operating the fitting of the given function.

    Args:
        func (``Callable[[np.ndarray], np.ndarray]``): Function to be optimised. If `do_grad=True`,
            function is expected to return `Tuple[np.ndarray, np.ndarray]` where first is the value
            of the function and the second is gradients with respect to the nuisance parameters.
        model_configuration (``~spey.base.model_config.ModelConfig``): Model configuration.
        do_grad (``bool``, default ``False``): boolean to set gradient on or off.
        hessian (``Optional[Callable[[np.ndarray], np.ndarray]]``, default ``None``): hessian of the function
            with respect to variational parameters. Currently not used.
        initial_parameters (``Optional[np.ndarray]``, default ``None``): Initial guess for the variational
            parameters.
        bounds (``Optional[List[Tuple[float, float]]]``, default ``None``): Bounds for the variational parameters
            each item has to have a lower and upper bound definition `(<lower>, <upper>)` in case of open boundary
            input can be `None`. Number of boundaries has to match with the number of parameters. They will match
            with respect to the index within the list.
        fixed_poi_value (``Optional[float]``, default ``None``): If `float` the poi index will be fixed.
        logpdf (``Optional[Callable[[List[float]], float]]``, default ``None``): If provided, log-probability
            distribution will be computed once the parameters are fitted. If not,  the result of the ``func`` input
            will be returned.
        constraints (``Optional[List[Dict]]``, default ``None``): Constraints of the model. see scipy for details.
        options (``Dict``): extra options for the optimiser. see scipy minimiser for details.

    Returns:
        ``Tuple[float, np.ndarray]``:
        value and the parameters. log-probability value will be returned if `logpdf` argument is provided.
    """

    init_pars = [*(initial_parameters or model_configuration.suggested_init)]
    par_bounds = [*(bounds or model_configuration.fixed_poi_bounds(fixed_poi_value))]

    def make_constraint(index: int, value: float) -> Callable[[np.ndarray], float]:
        def func(vector: np.ndarray) -> float:
            return vector[index] - value

        return func

    constraints = [] if constraints is None else constraints
    if fixed_poi_value is not None:
        init_pars[model_configuration.poi_index] = fixed_poi_value
        log.debug(
            f"Fixing POI index at: {model_configuration.poi_index} to value {fixed_poi_value}"
        )
        constraints.append(
            {
                "type": "eq",
                "fun": make_constraint(model_configuration.poi_index, fixed_poi_value),
            }
        )

    if model_configuration.suggested_fixed is not None:
        for idx, isfixed in enumerate(model_configuration.suggested_fixed):
            if isfixed:
                log.debug(f"Constraining {idx} to value {init_pars[idx]}")
                constraints.append(
                    {
                        "type": "eq",
                        "fun": make_constraint(idx, init_pars[idx]),
                    }
                )

    options.update({"poi_index": model_configuration.poi_index})

    fun, x = minimize(
        func=func,
        init_pars=init_pars,
        do_grad=do_grad,
        hessian=hessian,
        bounds=par_bounds,
        constraints=constraints,
        **options,
    )

    return fun if logpdf is None else logpdf(x), x



"""
Test statistic distributions

Similar implementation can be found in https://github.com/scikit-hep/pyhf/blob/main/src/pyhf/infer/calculators.py
"""

import numpy as np
from scipy.stats import multivariate_normal

__all__ = ["AsymptoticTestStatisticsDistribution", "EmpricTestStatisticsDistribution"]


def __dir__():
    return __all__


class AsymptoticTestStatisticsDistribution:
    """
    The distribution the test statistic in the asymptotic case.

    Args:
        shift (``float``): The shift in test statistic distribution.
    """

    __slots__ = ["shift", "cutoff"]

    def __init__(self, shift: float, cutoff: float = -np.inf):
        self.shift = shift
        self.cutoff = cutoff

    def pvalue(self, value: float) -> float:
        r"""
        Compute the p-value for a given value of the test statistic.

        Args:
            value (``float``): The test statistic value.

        Returns:
            ``float``:
            p-value
        """
        return_value = multivariate_normal.cdf(self.shift - value)
        return return_value if return_value >= self.cutoff else np.nan

    def expected_value(self, nsigma: float) -> float:
        """
        Compute expected value of the test statistic with respec to the level of standard deviations

        Args:
            nsigma (``float``): level of standard deviations.

        Returns:
            ``float``:
            expected value of test statistic.
        """
        tot = nsigma + self.shift
        return tot if tot > self.cutoff else self.cutoff


class EmpricTestStatisticsDistribution:
    """
    Create emprical distribution. The p-values are computed from sampled distribution.

    Args:
        samples (``np.ndarray``): input samples
    """

    __slots__ = ["samples"]

    def __init__(self, samples: np.ndarray) -> None:
        self.samples = samples

    def pvalue(self, value: float) -> float:
        """
        Compute the p-value for a given test statistic.

        Args:
            value (``float``): test statistic value

        Returns:
            ``float``:
            p-value
        """

        return np.sum(np.where(self.samples >= value, 1.0, 0.0)) / self.samples.shape[0]

    def expected_value(self, nsigma: float) -> float:
        """
        Compute expected value of the test statistic with respec to the level of standard deviations

        Args:
            nsigma (``float``): level of standard deviations.

        Returns:
            ``float``:
            expected value of test statistic.
        """
        return np.percentile(
            self.samples, multivariate_normal.cdf(nsigma) * 100.0, method="linear"
        )



"""Module specific exceptions"""

__all__ = [
    "FrozenInstanceError",
    "AnalysisQueryError",
    "NegativeExpectedYields",
    "UnknownCrossSection",
    "UnknownTestStatistics",
    "InvalidInput",
    "PluginError",
    "MethodNotAvailable",
]


class FrozenInstanceError(Exception):
    """Frozen instance exception"""

    def __init__(self, message="This class has been frozen."):
        super().__init__(message)


class AnalysisQueryError(Exception):
    """Analysis query exception"""

    def __init__(self, message="This analysis has not been found."):
        super().__init__(message)


class NegativeExpectedYields(Exception):
    """Negative expected yields exception"""

    def __init__(self, message="Negative expected yields has been found."):
        super().__init__(message)


class UnknownCrossSection(Exception):
    """Unknown cross-section exception"""

    def __init__(self, message="Please initialise cross section value."):
        super().__init__(message)


class UnknownTestStatistics(Exception):
    """Unknown test statistics exception"""

    def __init__(self, message="Unknown test statistics."):
        super().__init__(message)


class InvalidInput(Exception):
    """Invalid input exception"""

    def __init__(self, message="Unknown input type."):
        super().__init__(message)


class PluginError(Exception):
    """Invalid plugin exception"""


class MethodNotAvailable(Exception):
    """If the method is not available for a given backend"""


class CanNotFindRoots(Exception):
    """Unable to find roots of the function"""


class UnknownComputer(Exception):
    """Unknown computation base"""


class CalculatorNotAvailable(Exception):
    """Unavailable calculator Exception"""


class CombinerNotAvailable(Exception):
    """Unavailable combination routine exception"""


class DistributionError(Exception):
    """Unknown Distribution"""


class AsimovTestStatZero(Exception):
    """Asimov Test Statistic is zero"""

    def __init__(
        self,
        message="Asimov test statistic is zero. "
        "Note: Asimov test statistic of zero indicates a "
        "lack of evidence for a signal or deviation from a null hypothesis.",
    ):
        super().__init__(message)



"""Test default grad and hessian"""

import numpy as np
from spey.math import value_and_grad, hessian
import spey


def test_uncorrelated_background():
    """tester for uncorrelated background model"""

    pdf_wrapper = spey.get_backend("default_pdf.uncorrelated_background")

    data = [36, 33]
    signal_yields = [12.0, 15.0]
    background_yields = [50.0, 48.0]
    background_unc = [12.0, 16.0]

    stat_model = pdf_wrapper(
        signal_yields=signal_yields,
        background_yields=background_yields,
        data=data,
        absolute_uncertainties=background_unc,
        analysis="multi_bin",
        xsection=0.123,
    )
    hess = hessian(stat_model)([1.0, 1.0])
    nll, grad = value_and_grad(stat_model)([1.0, 1.0])
    nll_apri, grad_apri = value_and_grad(
        stat_model, expected=spey.ExpectationType.apriori
    )([1.0, 1.0])
    hess_apri = hessian(stat_model, expected=spey.ExpectationType.apriori)([1.0, 1.0])

    hess_data = hessian(stat_model, expected=spey.ExpectationType.apriori, data=[22, 34])(
        [1.0, 1.0]
    )
    nll_dat, grad_dat = value_and_grad(
        stat_model, expected=spey.ExpectationType.apriori, data=[22, 34]
    )([1.0, 1.0])

    assert np.allclose(
        hess, np.array([[2.13638959, 2.21570381], [2.21570381, 4.30030563]])
    ), "Hessian is wrong"
    assert np.isclose(nll, 37.47391613937222), "NLL is wrong"
    assert np.allclose(grad, np.array([14.89633938, 17.47861786])), "Gradient is wrong"
    assert np.isclose(nll_apri, 20.052816087791097), "NLL apriori is wrong"
    assert np.allclose(
        grad_apri, np.array([9.77796784, 12.1703729])
    ), "apriori Gradient is wrong"
    assert np.allclose(
        hess_apri, np.array([[3.04532025, 3.16068638], [3.16068638, 5.28374358]])
    ), "Hessian apriori is wrong"
    assert np.allclose(
        hess_data, np.array([[1.80428957, 1.88600725], [1.88600725, 3.97317276]])
    ), "Hessian data is wrong"
    assert np.isclose(nll_dat, 49.63922692607177), "NLL data is wrong"
    assert np.allclose(
        grad_dat, np.array([16.97673623, 19.54635648])
    ), "data Gradient is wrong"



"""
Statistical Model combiner class: this class combines likelihoods 
of different statistical models for hypothesis testing
"""

import warnings
from typing import Dict, Iterator, List, Optional, Text, Tuple, Union

import numpy as np

from spey.base.hypotest_base import HypothesisTestingBase
from spey.base.model_config import ModelConfig
from spey.interface.statistical_model import StatisticalModel
from spey.optimizer.core import fit
from spey.system.exceptions import AnalysisQueryError, NegativeExpectedYields
from spey.utils import ExpectationType

__all__ = ["UnCorrStatisticsCombiner"]


class UnCorrStatisticsCombiner(HypothesisTestingBase):
    """
    Module to combine **uncorrelated** statistical models. It takes serries of
    :obj:`~spey.StatisticalModel` object as input. These statistical models does not
    need to have same backend. However, this class assumes that all the input
    statistical model's are completely independent from each other.

    .. warning::

        :obj:`~spey.UnCorrStatisticsCombiner` assumes that all input are uncorrelated and
        non of the statistical models posesses the same set of nuisance parameters. Thus
        each statistical model is optimised independently.

    Args:
        input arguments (:obj:`~spey.StatisticalModel`): Uncorrelated statistical models
        ntoys (``int``, default ``1000``): Number of toy samples for hypothesis testing.
          (Only used for toy-based hypothesis testing)

    Raises:
        :obj:`~spey.system.exceptions.AnalysisQueryError`: If multiple :class:`~spey.StatisticalModel`
          has the same :attr:`~spey.StatisticalModel.analysis` attribute.
        :obj:`TypeError`: If the input type is not :class:`~spey.StatisticalModel`.
    """

    __slots__ = ["_statistical_models"]

    def __init__(self, *args, ntoys: int = 1000):
        super().__init__(ntoys=ntoys)
        self._statistical_models = []
        for arg in args:
            self.append(arg)

    def append(self, statistical_model: StatisticalModel) -> None:
        """
        Append new independent :class:`~spey.StatisticalModel` to the stack.

        Args:
            statistical_model (:class:`~spey.StatisticalModel`): new statistical model
              to be added to the stack.

        Raises:
            :obj:`~spey.system.exceptions.AnalysisQueryError`: If multiple :class:`~spey.StatisticalModel`
              has the same :attr:`~spey.StatisticalModel.analysis` attribute.
            :obj:`TypeError`: If the input type is not :class:`~spey.StatisticalModel`.
        """
        if isinstance(statistical_model, StatisticalModel):
            if statistical_model.analysis in self.analyses:
                raise AnalysisQueryError(f"{statistical_model.analysis} already exists.")
            self._statistical_models.append(statistical_model)
        else:
            raise TypeError(f"Can not append type {type(statistical_model)}.")

    def remove(self, analysis: Text) -> None:
        """
        Remove an analysis from the stack.

        Args:
            analysis (``Text``): unique identifier of the analysis to be removed.

        Raises:
            :obj:`~spey.system.exceptions.AnalysisQueryError`: If the unique identifier does not match
              any of the statistical models in the stack.
        """
        to_remove = None
        for name, model in self.items():
            if name == analysis:
                to_remove = model
        if to_remove is None:
            raise AnalysisQueryError(f"'{analysis}' is not among the analyses.")
        self._statistical_models.remove(to_remove)

    @property
    def statistical_models(self) -> Tuple[StatisticalModel]:
        """
        Accessor to the statistical model stack

        Returns:
            ``Tuple[StatisticalModel]``:
            statistical model stack as an ntuple.
        """
        return tuple(self._statistical_models)

    @property
    def analyses(self) -> List[Text]:
        """
        List of the unique identifiers of the statistical models within the stack.

        Returns:
            ``List[Text]``:
            List of analysis names.
        """
        return [model.analysis for model in self]

    @property
    def minimum_poi(self) -> float:
        r"""
        Find the maximum minimum poi, :math:`\mu` value within the analysis stack.

        Returns:
            ``float``:
            minimum poi value that the combined stack can take.
        """
        return max(model.backend.config().minimum_poi for model in self)

    @property
    def is_alive(self) -> bool:
        """
        Returns true if there is at least one statistical model with at least
        one non zero bin with signal yield.

        Returns:
            ``bool``
        """
        return any(model.is_alive for model in self)

    @property
    def is_asymptotic_calculator_available(self) -> bool:
        """Check if Asymptotic calculator is available for the backend"""
        return all(model.is_asymptotic_calculator_available for model in self)

    @property
    def is_toy_calculator_available(self) -> bool:
        """Check if Toy calculator is available for the backend"""
        return False

    @property
    def is_chi_square_calculator_available(self) -> bool:
        r"""Check if :math:`\chi^2` calculator is available for the backend"""
        return all(model.is_chi_square_calculator_available for model in self)

    def __getitem__(self, item: Union[Text, int]) -> StatisticalModel:
        """Retrieve a statistical model"""
        if isinstance(item, int):
            if item < len(self):
                return self.statistical_models[item]
            raise AnalysisQueryError(
                "Request exceeds number of statistical models available."
            )
        if isinstance(item, slice):
            return self.statistical_models[item]

        for model in self:
            if model.analysis == item:
                return model
        raise AnalysisQueryError(f"'{item}' is not among the analyses.")

    def __iter__(self) -> StatisticalModel:
        """Iterate over statistical models"""
        yield from self._statistical_models

    def __len__(self):
        """Number of statistical models within the stack"""
        return len(self._statistical_models)

    def items(self) -> Iterator[Tuple[Text, StatisticalModel]]:
        """
        Returns a generator that returns analysis name and :obj:`~spey.StatisticalModel`
        every iteration.

        Returns:
            ``Iterator[Tuple[Text, StatisticalModel]]``
        """
        return ((model.analysis, model) for model in self)

    def find_most_sensitive(self) -> StatisticalModel:
        """
        If the cross sections are defined, finds the statistical model with minimum prefit
        excluded cross section value. See :attr:`~spey.StatisticalModel.s95exp`.

        Returns:
            :obj:`~spey.StatisticalModel`:
            Statistical model with minimum prefit excluded cross section value.
        """
        return self[np.argmin([model.s95exp for model in self])]

    def likelihood(
        self,
        poi_test: float = 1.0,
        expected: ExpectationType = ExpectationType.observed,
        return_nll: bool = True,
        data: Optional[Dict[Text, List[float]]] = None,
        statistical_model_options: Optional[Dict[Text, Dict]] = None,
        **kwargs,
    ) -> float:
        r"""
        Compute the likelihood of the statistical model stack at a fixed parameter of interest

        Args:
            poi_test (``float``, default ``1.0``): parameter of interest or signal strength,
              :math:`\mu`.
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

            return_nll (``bool``, default ``True``): If ``True``, returns negative log-likelihood value.
              if ``False`` returns likelihood value.
            data (``Dict[Text, List[float]]``, default ``None``): Input data to be used for fit. It needs
              to be given as a dictionary where the key argument is the name of the analysis and item is
              the statistical model specific data.
            statistical_model_options (``Optional[Dict[Text, Dict]]``, default ``None``): backend specific
              options. The dictionary key needs to be the backend name and the item needs to be the dictionary
              holding the keyword arguments specific to that particular backend.

              .. code-block:: python3

                >>> statistical_model_options = {"default_pdf.uncorrelated_background" : {"init_pars" : [1., 3., 4.]}}

            kwargs: keyword arguments for the optimiser.

        Returns:
            ``float``:
            value of the likelihood of the stacked statistical model.
        """
        statistical_model_options = statistical_model_options or {}
        data = data or {}

        nll = 0.0
        for statistical_model in self:

            current_kwargs = {}
            current_kwargs.update(
                statistical_model_options.get(str(statistical_model.backend_type), {})
            )
            current_data = data.get(statistical_model.analysis, None)

            try:
                nll += statistical_model.likelihood(
                    poi_test=poi_test,
                    expected=expected,
                    data=current_data,
                    **current_kwargs,
                    **kwargs,
                )
            except NegativeExpectedYields as err:
                warnings.warn(
                    err.args[0] + f"\nSetting NLL({poi_test:.3f}) = nan",
                    category=RuntimeWarning,
                )
                nll = np.nan

            if np.isnan(nll):
                break

        return nll if return_nll or np.isnan(nll) else np.exp(-nll)

    def generate_asimov_data(
        self,
        expected: ExpectationType = ExpectationType.observed,
        test_statistic: Text = "qtilde",
        statistical_model_options: Optional[Dict[Text, Dict]] = None,
        **kwargs,
    ) -> Dict[Text, List[float]]:
        r"""
        Generate Asimov data for the statistical model. This function calls
        :func:`~spey.StatisticalModel.generate_asimov_data` function for each statistical model with
        appropriate ``statistical_model_options`` and generates Asimov data for each statistical
        model independently.

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

            test_statistic (``Text``, default ``"qtilde"``): test statistics.

              * ``'qtilde'``: (default) performs the calculation using the alternative test statistic,
                :math:`\tilde{q}_{\mu}`, see eq. (62) of :xref:`1007.1727`
                (:func:`~spey.hypothesis_testing.test_statistics.qmu_tilde`).

                .. warning::

                    Note that this assumes that :math:`\hat\mu\geq0`, hence ``allow_negative_signal``
                    assumed to be ``False``. If this function has been executed by user, ``spey``
                    assumes that this is taken care of throughout the external code consistently.
                    Whilst computing p-values or upper limit on :math:`\mu` through ``spey`` this
                    is taken care of automatically in the backend.

              * ``'q'``: performs the calculation using the test statistic :math:`q_{\mu}`, see
                eq. (54) of :xref:`1007.1727` (:func:`~spey.hypothesis_testing.test_statistics.qmu`).
              * ``'q0'``: performs the calculation using the discovery test statistic, see eq. (47)
                of :xref:`1007.1727` :math:`q_{0}` (:func:`~spey.hypothesis_testing.test_statistics.q0`).

            statistical_model_options (``Dict[Text, Dict]``, default ``None``): backend specific
              options. The dictionary key needs to be the backend name and the item needs to be the dictionary
              holding the keyword arguments specific to that particular backend.

              .. code-block:: python3

                >>> statistical_model_options = {"default_pdf.uncorrelated_background" : {"init_pars" : [1., 3., 4.]}}

            kwargs: keyword arguments for the optimiser.

        Returns:
            ``Dict[Text, List[float]]``:
            Returns a dictionary for data specific to each analysis. keywords will be analysis names
            and the items are data.
        """
        statistical_model_options = statistical_model_options or {}

        data = {}
        for statistical_model in self:
            current_kwargs = {}
            current_kwargs.update(
                statistical_model_options.get(str(statistical_model.backend_type), {})
            )

            data.update(
                {
                    statistical_model.analysis: statistical_model.generate_asimov_data(
                        expected=expected,
                        test_statistic=test_statistic,
                        **current_kwargs,
                        **kwargs,
                    )
                }
            )

        return data

    def asimov_likelihood(
        self,
        poi_test: float = 1.0,
        expected: ExpectationType = ExpectationType.observed,
        return_nll: bool = True,
        test_statistics: Text = "qtilde",
        statistical_model_options: Optional[Dict[Text, Dict]] = None,
        **kwargs,
    ) -> float:
        r"""
        Compute likelihood of the statistical model stack generated with the Asimov data.

        Args:
            poi_test (``float``, default ``1.0``): parameter of interest, :math:`\mu`.
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

            return_nll (``bool``, default ``True``): If ``True``, returns negative log-likelihood value.
              if ``False`` returns likelihood value.
            test_statistics (``Text``, default ``"qtilde"``): test statistics.

              * ``'qtilde'``: (default) performs the calculation using the alternative test statistic,
                :math:`\tilde{q}_{\mu}`, see eq. (62) of :xref:`1007.1727`
                (:func:`~spey.hypothesis_testing.test_statistics.qmu_tilde`).

                .. warning::

                    Note that this assumes that :math:`\hat\mu\geq0`, hence ``allow_negative_signal``
                    assumed to be ``False``. If this function has been executed by user, ``spey``
                    assumes that this is taken care of throughout the external code consistently.
                    Whilst computing p-values or upper limit on :math:`\mu` through ``spey`` this
                    is taken care of automatically in the backend.

              * ``'q'``: performs the calculation using the test statistic :math:`q_{\mu}`, see
                eq. (54) of :xref:`1007.1727` (:func:`~spey.hypothesis_testing.test_statistics.qmu`).
              * ``'q0'``: performs the calculation using the discovery test statistic, see eq. (47)
                of :xref:`1007.1727` :math:`q_{0}` (:func:`~spey.hypothesis_testing.test_statistics.q0`).

            statistical_model_options (``Dict[Text, Dict]``, default ``None``): backend specific
              options. The dictionary key needs to be the backend name and the item needs to be the dictionary
              holding the keyword arguments specific to that particular backend.

              .. code-block:: python3

                >>> statistical_model_options = {"default_pdf.uncorrelated_background" : {"init_pars" : [1., 3., 4.]}}

            kwargs: keyword arguments for the optimiser.

        Returns:
            ``float``:
            likelihood computed for asimov data
        """
        return self.likelihood(
            poi_test=poi_test,
            expected=expected,
            return_nll=return_nll,
            data=self.generate_asimov_data(
                expected=expected,
                test_statistic=test_statistics,
                statistical_model_options=statistical_model_options,
                **kwargs,
            ),
            statistical_model_options=statistical_model_options,
            **kwargs,
        )

    def maximize_likelihood(
        self,
        return_nll: bool = True,
        expected: ExpectationType = ExpectationType.observed,
        allow_negative_signal: bool = True,
        data: Optional[Dict[Text, List[float]]] = None,
        initial_muhat_value: Optional[float] = None,
        par_bounds: Optional[List[Tuple[float, float]]] = None,
        statistical_model_options: Optional[Dict[Text, Dict]] = None,
        **optimiser_options,
    ) -> Tuple[float, float]:
        r"""
        Find the maximum of the likelihood.

        Args:
            return_nll (``bool``, default ``True``): If  ``True``, returns negative log-likelihood value.
              if ``False`` returns likelihood value.
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

            allow_negative_signal (``bool``, default ``True``): If ``True`` :math:`\hat\mu`
              value will be allowed to be negative.
            data (``Dict[Text, List[float]]``, default ``None``): Input data to be used for fit. It needs
              to be given as a dictionary where the key argument is the name of the analysis and item is
              the statistical model specific data.
            initial_muhat_value (``float``, default ``None``): Initialisation for the :math:`\hat\mu` for
              the optimiser. If ``None`` the initial value will be estimated by weighted combination of
              :math:`\hat\mu_i` where :math:`i` stands for the statistical models within the stack.

              .. math::

                    \mu_{\rm init} = \frac{1}{\sum_i \sigma_{\hat\mu, i}^2} \sum_i \frac{\hat\mu_i}{\sigma_{\hat\mu, i}^2}

              where the value of :math:`\sigma_{\hat\mu}` has been estimated by
              :func:`~spey.base.hypotest_base.HypothesisTestingBase.sigma_mu` function.
            par_bounds (``List[Tuple[float, float]]``, default  ``None``): parameter bounds for
              the optimiser.
            statistical_model_options (``Dict[Text, Dict]``, default ``None``): backend specific
              options. The dictionary key needs to be the backend name and the item needs to be the dictionary
              holding the keyword arguments specific to that particular backend.

              .. code-block:: python3

                >>> statistical_model_options = {"default_pdf.uncorrelated_background" : {"init_pars" : [1., 3., 4.]}}

            kwargs: keyword arguments for the optimiser.

        Returns:
            ``Tuple[float, float]``:
            :math:`\hat\mu` value and maximum value of the likelihood.
        """
        statistical_model_options = statistical_model_options or {}
        data = data or {}

        # muhat initial value estimation in gaussian limit
        mu_init = initial_muhat_value or 0.0
        if initial_muhat_value is None:
            _mu, _sigma_mu = np.zeros(len(self)), np.ones(len(self))
            for idx, stat_model in enumerate(self):

                current_kwargs = {}
                current_kwargs.update(
                    statistical_model_options.get(str(stat_model.backend_type), {})
                )

                _mu[idx] = stat_model.maximize_likelihood(
                    expected=expected, **current_kwargs, **optimiser_options
                )[0]
                _sigma_mu[idx] = stat_model.sigma_mu(
                    poi_test=_mu[idx],
                    expected=expected,
                    **current_kwargs,
                    **optimiser_options,
                )
            norm = np.sum(np.power(_sigma_mu, -2))
            mu_init = np.true_divide(1.0, norm) * np.sum(
                np.true_divide(_mu, np.square(_sigma_mu))
            )

        config: ModelConfig = ModelConfig(
            poi_index=0,
            minimum_poi=self.minimum_poi,
            suggested_init=[float(mu_init)],
            suggested_bounds=(
                par_bounds
                or [
                    (
                        self.minimum_poi if allow_negative_signal else 0.0,
                        max(10.0, mu_init),
                    )
                ]
            ),
        )

        def twice_nll(poi_test: Union[float, np.ndarray]) -> float:
            """Function to compute twice negative log-likelihood for a given poi test"""
            return 2.0 * self.likelihood(
                poi_test if isinstance(poi_test, float) else poi_test[0],
                expected=expected,
                return_nll=True,
                data=data,
                statistical_model_options=statistical_model_options,
                **optimiser_options,
            )

        twice_nll, fit_params = fit(
            func=twice_nll,
            model_configuration=config,
            **optimiser_options,
        )

        return fit_params[0], twice_nll / 2.0 if return_nll else np.exp(-twice_nll / 2.0)

    def maximize_asimov_likelihood(
        self,
        return_nll: bool = True,
        expected: ExpectationType = ExpectationType.observed,
        test_statistics: Text = "qtilde",
        initial_muhat_value: Optional[float] = None,
        par_bounds: Optional[List[Tuple[float, float]]] = None,
        statistical_model_options: Optional[Dict[Text, Dict]] = None,
        **optimiser_options,
    ) -> Tuple[float, float]:
        r"""
        Find the maximum of the likelihood which computed with respect to Asimov data.

        Args:
            return_nll (``bool``, default ``True``): If ``True``, returns negative log-likelihood value.
              if ``False`` returns likelihood value.
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

            test_statistics (``Text``, default ``"qtilde"``): test statistic.

              * ``'qtilde'``: (default) performs the calculation using the alternative test statistic,
                :math:`\tilde{q}_{\mu}`, see eq. (62) of :xref:`1007.1727`
                (:func:`~spey.hypothesis_testing.test_statistics.qmu_tilde`).

                .. warning::

                    Note that this assumes that :math:`\hat\mu\geq0`, hence :obj:`allow_negative_signal`
                    assumed to be :obj:`False`. If this function has been executed by user, :obj:`spey`
                    assumes that this is taken care of throughout the external code consistently.
                    Whilst computing p-values or upper limit on :math:`\mu` through :obj:`spey` this
                    is taken care of automatically in the backend.

              * ``'q'``: performs the calculation using the test statistic :math:`q_{\mu}`, see
                eq. (54) of :xref:`1007.1727` (:func:`~spey.hypothesis_testing.test_statistics.qmu`).
              * ``'q0'``: performs the calculation using the discovery test statistic, see eq. (47)
                of :xref:`1007.1727` :math:`q_{0}` (:func:`~spey.hypothesis_testing.test_statistics.q0`).

            initial_muhat_value (``float``, default ``None``): Initial value for :math:`\hat\mu`.
            par_bounds (``List[Tuple[float, float]]``, default ``None``): parameter bounds for
              the optimiser.
            statistical_model_options (``Dict[Text, Dict]``, default ``None``): backend specific
              options. The dictionary key needs to be the backend name and the item needs to be the dictionary
              holding the keyword arguments specific to that particular backend.

              .. code-block:: python3

                >>> statistical_model_options = {"default_pdf.uncorrelated_background" : {"init_pars" : [1., 3., 4.]}}

            kwargs: keyword arguments for the optimiser.

        Returns:
            ``Tuple[float, float]``:
            :math:`\hat\mu` value and maximum value of the likelihood.
        """
        allow_negative_signal: bool = True if test_statistics in ["q", "qmu"] else False

        config: ModelConfig = ModelConfig(
            poi_index=0,
            minimum_poi=self.minimum_poi,
            suggested_init=[initial_muhat_value or 0.0],
            suggested_bounds=(
                par_bounds or [(self.minimum_poi if allow_negative_signal else 0.0, 10.0)]
            ),
        )

        statistical_model_options = statistical_model_options or {}

        data = self.generate_asimov_data(
            expected=expected,
            test_statistic=test_statistics,
            statistical_model_options=statistical_model_options,
            **optimiser_options,
        )

        def twice_nll(poi_test: Union[float, np.ndarray]) -> float:
            """Function to compute twice negative log-likelihood for a given poi test"""
            return 2.0 * self.likelihood(
                poi_test if isinstance(poi_test, float) else poi_test[0],
                expected=expected,
                data=data,
                statistical_model_options=statistical_model_options,
                **optimiser_options,
            )

        twice_nll, fit_params = fit(
            func=twice_nll,
            model_configuration=config,
            **optimiser_options,
        )

        return fit_params[0], twice_nll / 2.0 if return_nll else np.exp(-twice_nll / 2.0)



"""Autograd based differentiable distribution classes"""

import logging
import warnings
from typing import Any, Callable, Dict, List, Literal, Text, Union

import autograd.numpy as np
from autograd.scipy.special import gammaln
from autograd.scipy.stats.poisson import logpmf
from scipy.stats import multivariate_normal, norm, poisson

from spey.system.exceptions import DistributionError

# pylint: disable=E1101, W1203, E1121

log = logging.getLogger("Spey")

__all__ = ["Poisson", "Normal", "MultivariateNormal", "MainModel", "ConstraintModel"]


def __dir__():
    return __all__


class Poisson:
    """Poisson distribution"""

    def __init__(self, loc: np.ndarray):
        # ! Clip for numeric stability, poisson can not take negative values
        self.loc = np.clip(loc, 1e-20, None)

    def expected_data(self) -> np.ndarray:
        """The expectation value of the Poisson distribution."""
        return np.array(self.loc)

    def sample(self, sample_size: int) -> np.ndarray:
        """Generate samples"""
        shape = [sample_size]
        if isinstance(self.loc, np.ndarray):
            shape += [len(self.loc)]
        return poisson(self.loc).rvs(size=shape)

    def log_prob(self, value: np.ndarray) -> np.ndarray:
        """Compute log-probability"""
        # for code efficiency
        if np.array(value).dtype in [np.int32, np.int16, np.int64]:
            return logpmf(value, self.loc).astype(np.float64)

        return (value * np.log(self.loc) - self.loc - gammaln(value + 1.0)).astype(
            np.float64
        )


class Normal:
    """
    Normal distribution

    Args:
        loc (``np.ndarray``): Mean of the distribution.
        scale (``np.ndarray``): standard deviation.
        weight (``Callable[[np.ndarray], float]`` or ``float``, default ``1.0``): weight of
          the distribution.
        domain (``slice``, default ``slice(None, None)``): set of parameters to be used within
          the distribution.
    """

    def __init__(
        self,
        loc: np.ndarray,
        scale: np.ndarray,
        weight: Union[Callable[[np.ndarray], float], float] = 1.0,
        domain: slice = slice(None, None),
    ):
        self.loc = loc
        self.weight = weight if callable(weight) else lambda pars: weight
        """Weight of the distribution"""
        self.domain = domain
        """Which parameters should be used during the computation of the pdf"""

        if callable(scale):
            self.scale = scale
        else:
            self.scale = lambda pars: scale

    def expected_data(self) -> np.ndarray:
        """The expectation value of the Normal distribution."""
        return np.array(self.loc)

    def sample(self, value: np.ndarray, sample_size: int) -> np.ndarray:
        """Generate samples"""
        shape = [sample_size]
        if isinstance(self.loc, np.ndarray):
            shape += [len(self.loc)]

        return norm(self.loc, self.scale(value[self.domain])).rvs(size=shape)

    def log_prob(self, value: float) -> np.ndarray:
        """Compute log-probability"""
        return self.weight(value) * (
            -np.log(self.scale(value[self.domain]))
            - 0.5 * np.log(2.0 * np.pi)
            - 0.5
            * np.square(
                np.divide(value[self.domain] - self.loc, self.scale(value[self.domain]))
            )
        ).astype(np.float64)


class MultivariateNormal:
    """
    Multivariate normal distribution

    Args:
        mean (``np.ndarray``): Mean of the distribution.
        cov (``np.ndarray``): Symmetric positive (semi)definite
          covariance matrix of the distribution.
        weight (``Callable[[np.ndarray], float]`` or ``float``, default ``1.0``): weight of
          the distribution.
        domain (``slice``, default ``slice(None, None)``): set of parameters to be used within
          the distribution.
    """

    def __init__(
        self,
        mean: np.ndarray,
        cov: np.ndarray,
        weight: Union[Callable[[np.ndarray], float], float] = 1.0,
        domain: slice = slice(None, None),
    ):
        self.mean = mean
        """Mean of the distribution."""
        self.cov = cov if callable(cov) else lambda val: cov
        """Symmetric positive (semi)definite covariance matrix of the distribution."""
        self.weight = weight if callable(weight) else lambda pars: weight
        """Weight of the distribution"""
        self.domain = domain
        """Which parameters should be used during the computation of the pdf"""

        # ! the min determinant value of the covariance matrix is artificially set
        # ! to 1e-10 this might cause problems in the future!!!

        if callable(cov):
            self._inv_cov = lambda val: np.linalg.inv(cov(val))
            self._det_cov = lambda val: np.clip(np.linalg.det(cov(val)), 1e-20, None)
        else:
            # for code efficiency
            inv = np.linalg.inv(cov)
            det = np.linalg.det(cov)
            if det <= 0.0:
                warnings.warn(
                    "det(rho) <= 0, this might cause numeric problems. "
                    "The value of the determinant will be limited to 1e-20. "
                    "This might be due to non-positive definite correlation matrix input."
                )
                det = np.clip(det, 1e-20, None)
            self._inv_cov = lambda val: inv
            self._det_cov = lambda val: det

    def expected_data(self) -> np.ndarray:
        """The expectation value of the Multivariate Normal distribution."""
        return self.mean

    def sample(self, value: np.ndarray, sample_size: int) -> np.ndarray:
        """Generate samples"""
        return multivariate_normal(self.mean, self.cov(value[self.domain])).rvs(
            size=(sample_size,)
        )

    def log_prob(self, value: np.ndarray) -> np.ndarray:
        """Compute log-probability"""
        var = value[self.domain] - self.mean
        return self.weight(value) * (
            -0.5 * (var @ self._inv_cov(value[self.domain]) @ var)
            - 0.5
            * (
                len(value[self.domain]) * np.log(2.0 * np.pi)
                + np.log(self._det_cov(value[self.domain]))
            )
        ).astype(np.float64)


class MainModel:
    """
    Main statistical model, modelled as Poisson distribution which has a
    variable lambda.

    Args:
        loc (``Callable[[np.ndarray], np.ndarray]``): callable function that represents
          lambda values of poisson distribution. It takes nuisance parameters as input.
    """

    def __init__(
        self,
        loc: Callable[[np.ndarray], np.ndarray],
        cov: np.ndarray = None,
        pdf_type: Literal["poiss", "gauss", "multivariategauss"] = "poiss",
    ):
        self.pdf_type = pdf_type
        """Type of the PDF"""
        if pdf_type == "poiss":
            self._pdf = lambda pars: Poisson(loc(pars))
        elif pdf_type == "gauss" and cov is not None:
            self._pdf = lambda pars: Normal(loc=loc(pars), scale=cov)
        elif pdf_type == "multivariategauss" and cov is not None:
            self._pdf = lambda pars: MultivariateNormal(mean=loc(pars), cov=cov)
        else:
            raise DistributionError("Unknown pdf type or associated input.")

    def expected_data(self, pars: np.ndarray) -> np.ndarray:
        """The expectation value of the main model."""
        return self._pdf(pars).expected_data()

    def sample(self, pars: np.ndarray, sample_size: int) -> np.ndarray:
        r"""
        Generate samples

        Args:
            pars (``np.ndarray``): parameter of interest and nuisance parameters
              :math:`\mu` and :math:`\theta` combined.
            sample_size (``int``): size of the sample to return

        Returns:
            ``np.ndarray``:
            sampled data
        """
        if self.pdf_type == "poiss":
            return self._pdf(pars).sample(sample_size)

        return self._pdf(pars).sample(pars, sample_size)

    def log_prob(self, pars: np.ndarray, data: np.ndarray) -> np.ndarray:
        r"""
        Compute log-probability

        Args:
            pars (``np.ndarray``): parameter of interest and nuisance parameters
              :math:`\mu` and :math:`\theta` combined.
            data (``np.ndarray``): actual data

        Returns:
            ``np.ndarray``:
            log-probability of the main model
        """
        return np.sum(self._pdf(pars).log_prob(data))


class ConstraintModel:
    """
    Constraint term modelled as a Gaussian distribution.

    Args:
        pdf_descriptions (``List[Dict[Text, Any]]``): description of the pdf component.
            Dictionary elements should contain two keywords

              * ``"distribution_type"`` (``Text``): ``"normal"`` or ``"multivariatenormal"``
              * ``"args"``: Input arguments for the distribution
              * ``"kwargs"``: Input keyword arguments for the distribution
    """

    def __init__(self, pdf_descriptions: List[Dict[Text, Any]]):
        self._pdfs = []
        distributions = {"normal": Normal, "multivariatenormal": MultivariateNormal}

        log.debug("Adding constraint terms:")
        for desc in pdf_descriptions:
            assert desc["distribution_type"].lower() in [
                "normal",
                "multivariatenormal",
            ], f"Unknown distribution type: {desc['distribution_type']}"
            log.debug(f"{desc}")
            self._pdfs.append(
                distributions[desc["distribution_type"]](
                    *desc.get("args", []), **desc.get("kwargs", {})
                )
            )

    def __len__(self):
        return len(self._pdfs)

    def expected_data(self) -> np.ndarray:
        """The expectation value of the constraint model."""
        if len(self) > 1:
            return np.hstack([pdf.expected_data() for pdf in self._pdfs])

        return self._pdfs[0].expected_data()

    def sample(self, pars: np.ndarray, sample_size: int) -> np.ndarray:
        r"""
        Generate samples

        Args:
            pars (``np.ndarray``): parameter of interest and nuisance parameters
              :math:`\mu` and :math:`\theta` combined.
            sample_size (``int``): size of the sample to return

        Returns:
            ``np.ndarray``:
            sampled data
        """
        if len(self) > 1:
            return np.hstack([pdf.sample(pars, sample_size) for pdf in self._pdfs])

        return self._pdfs[0].sample(pars, sample_size)

    def log_prob(self, pars: np.ndarray) -> np.ndarray:
        r"""
        Compute log-probability

        Args:
            pars (``np.ndarray``): parameter of interest and nuisance parameters
              :math:`\mu` and :math:`\theta` combined.
            data (``np.ndarray``): actual data

        Returns:
            ``np.ndarray``:
            log-probability of the main model
        """
        return np.sum([pdf.log_prob(pars) for pdf in self._pdfs])


class MixtureModel:
    """
    Generate probability distribution from combination of different distributions

    Args:
        args: Distributions

    .. warning::

        All distributions have been assumed to have same shape.
    """

    def __init__(self, *args):
        self.distributions = [d for d in args if hasattr(d, "log_prob")]

    def __iter__(self):
        yield from self.distributions

    def sample(self, sample_shape: int) -> np.ndarray:
        """Generate samples"""
        data = np.zeros(
            (sample_shape, self.distributions[-1].shape, len(self.distributions))
        )
        for idx, dist in enumerate(self):
            data[:, ..., idx] = dist.sample(sample_shape)
        random_idx = np.random.choice(
            np.arange(len(self.distributions)),
            size=(sample_shape,),
            p=[1.0 / len(self.distributions)] * len(self.distributions),
        )
        return data[np.arange(sample_shape), ..., random_idx]

    def log_prob(self, value: np.ndarray) -> np.ndarray:
        """Compute log-probability"""
        return np.sum([dist.log_prob(value) for dist in self.distributions])



"""Optimisation tools based on scipy"""

import logging
import warnings
from typing import Callable, Dict, List, Tuple

import numpy as np
import scipy

# pylint: disable=W1201, W1203, R0913

log = logging.getLogger("Spey")


def minimize(
    func: Callable[[np.ndarray], float],
    init_pars: List[float],
    do_grad: bool = False,
    hessian: Callable[[np.ndarray], np.ndarray] = None,
    bounds: List[Tuple[float, float]] = None,
    constraints: List[Dict] = None,
    **options,
) -> Tuple[float, np.ndarray]:
    """
    Minimize given function using scipy optimiser functionality

    Default options:
        method = "SLSQP"
        maxiter = 10000
        disp = False
        tol = 1e-6
        ntrials = 1

    :param func (`Callable[[np.ndarray], float]`): the objective function to be minimized.
    :param init_pars (`List[float]`): initial set of parameters
    :param do_grad (`bool`, default `False`): if true func is expected to return both objective and its gradient
    :param hessian (`Callable[[np.ndarray], np.ndarray]`, default `None`): hessian matrix of the objective function.
    :param bounds (`List[Tuple[float, float]]`, default `None`): Bounds on variables for Nelder-Mead, L-BFGS-B, TNC, SLSQP,
                                                                Powell, and trust-constr methods.
    :param constraints (`List[Dict]`, default `None`): Constraints definition. Only for COBYLA, SLSQP and trust-constr.
    :return `Tuple[float, np.ndarray]`: minimum value of the objective function and fit parameters
    """
    assert "poi_index" in list(options), "Please include `poi_index` in the options."

    method = options.pop("method", "SLSQP")
    tol = options.pop("tol", 1e-6)
    ntrials = max(options.pop("ntrials", 1), 1)
    poi_index = options.pop("poi_index")

    options.update({"maxiter": options.get("maxiter", 10000)})
    options.update({"disp": options.get("disp", False)})

    ntrial = 0
    while ntrial < ntrials:
        ntrial += 1
        with warnings.catch_warnings():
            warnings.filterwarnings(action="ignore", category=RuntimeWarning)
            opt = scipy.optimize.minimize(
                func,
                init_pars,
                method=method,
                jac=do_grad,
                hess=hessian,
                bounds=bounds,
                constraints=constraints,
                tol=tol,
                options=options,
            )
        if not opt.success and ntrial < ntrials:
            # Expand the boundaries of the statistical model dynamically to be able to converge
            # Note that it is important to respect the lower bounds especially since there might
            # be bounds that lead to negative yields in the statistical model, e.g. Asimov data,
            # background + mu * signal etc.
            log.warning("Optimiser has not been able to satisfy all the conditions:")
            log.warning(opt.message + " Expanding the bounds...")
            init_pars = opt.x
            for bdx, bound in enumerate(bounds):
                if bdx == poi_index and constraints is None:
                    bounds[bdx] = (bound[0], bound[1] * 2.0)
                else:
                    bounds[bdx] = (min(bound[0], bound[0] * 10.0), bound[1] * 10.0)
        elif opt.success:
            break

    if not opt.success:
        log.warning("Optimiser has not been able to satisfy all the conditions:")
        log.warning(opt.message)
        log.debug(f"func: {opt.fun}")
        log.debug(f"parameters: {opt.x}")
        if do_grad:
            log.debug(f"gradients: {func(opt.x)[-1]}")

    return opt.fun, opt.x



"""Configuration class for Statistical Models"""

import copy
from dataclasses import dataclass
from typing import List, Optional, Text, Tuple


@dataclass
class ModelConfig:
    r"""
    Container to hold certain properties of the backend and statistical model.
    This will ensure the consistency of the computation through out the package.

    Args:
        poi_index (:obj:`int`): index of the parameter of interest within the parameter list.
        minimum_poi (:obj:`float`): minimum value parameter of interest can take to ensure
          :math:`N^{\rm bkg}+\mu N^{\rm sig}\geq0`. This value can be set to :math:`-\infty`
          but note that optimiser will take it as a lower bound so such low value might effect
          the convergence of the optimisation algorithm especially for relatively flat objective
          surfaceses. Hence we suggest setting the minimum value to

            .. math::

                \mu_{\rm min} = - \min\left( \frac{N^{\rm bkg}_i}{N^{\rm sig}_i} \right)\ ,\ i\in {\rm bins}

        suggested_init (:obj:`List[float]`): suggested initial parameters for the optimiser.
        suggested_bounds (:obj:`List[Tuple[float, float]]`): suggested parameter bounds for the
          optimiser.

    Returns:
        :obj:~spey.base.model_config.ModelConfig:
        Model configuration container for optimiser.
    """

    poi_index: int
    """Index of the parameter of interest wihin the parameter list"""
    minimum_poi: float
    r"""
    minimum value parameter of interest can take to ensure :math:`N^{\rm bkg}+\mu N^{\rm sig}\geq0`.
    This value can be set to :math:`-\infty` but note that optimiser will take it as a lower bound so 
    such low value might effect the convergence of the optimisation algorithm especially for relatively
    flat objective surfaceses. Hence we suggest setting the minimum value to

    .. math::

        \mu_{\rm min} = - \min\left( \frac{N^{\rm bkg}_i}{N^{\rm sig}_i} \right)\ ,\ i\in {\rm bins}
    """
    suggested_init: List[float]
    """Suggested initialisation for parameters"""
    suggested_bounds: List[Tuple[float, float]]
    """Suggested upper and lower bounds for parameters"""
    parameter_names: Optional[List[Text]] = None
    """Names of the parameters"""
    suggested_fixed: Optional[List[bool]] = None
    """Suggested fixed values"""

    @property
    def npar(self) -> int:
        """Number of parameters"""
        return len(self.suggested_init)

    def fixed_poi_bounds(
        self, poi_value: Optional[float] = None
    ) -> List[Tuple[float, float]]:
        r"""
        Adjust the bounds for the parameter of interest for fixed POI fit.

        Args:
            poi_value (:obj:`Optional[float]`, default :obj:`None`): parameter of interest,
              :math:`\mu`.

        Returns:
            :obj:`List[Tuple[float, float]]`:
            Updated bounds.
        """
        if poi_value is None:
            return self.suggested_bounds
        bounds = copy.deepcopy(self.suggested_bounds)
        if any(b is None for b in bounds[self.poi_index]):
            return bounds
        if not bounds[self.poi_index][0] < poi_value < bounds[self.poi_index][1]:
            bounds[self.poi_index] = (
                self.minimum_poi if poi_value < 0.0 else 0.0,
                poi_value + 1,
            )
        return bounds

    def rescale_poi_bounds(
        self, allow_negative_signal: bool = True, poi_upper_bound: Optional[float] = None
    ) -> List[Tuple[float, float]]:
        r"""
        Rescale bounds for POI.

        Args:
            allow_negative_signal (:obj:`bool`, default :obj:`True`): If :obj:`True` :math:`\hat\mu`
              value will be allowed to be negative.
            poi_upper_bound (:obj:`float`, default :obj:`None`): Maximum value POI can take during
              optimisation.

        Returns:
            :obj:`List[Tuple[float, float]]`:
            Updated bounds.
        """
        bounds = copy.deepcopy(self.suggested_bounds)
        if poi_upper_bound:
            bounds[self.poi_index] = (
                self.minimum_poi if allow_negative_signal else 0.0,
                poi_upper_bound,
            )
        else:
            bounds[self.poi_index] = (
                self.minimum_poi if allow_negative_signal else 0.0,
                bounds[self.poi_index],
            )
        return bounds



import logging
from typing import Literal, Text

import requests
from semantic_version import Version

from spey._version import __version__

log = logging.getLogger("Spey")

__all__ = ["get_bibtex", "check_updates", "ConnectionError"]


def __dir__():
    return __all__


class ConnectionError(Exception):
    """No internet connection"""


def get_bibtex(
    home: Literal["inspire/arxiv", "inspire/doi", "doi", "zenodo"],
    identifier: Text,
    timeout: int = 5,
) -> Text:
    """
    Retreive BibTex information from InspireHEP, Zenodo or DOI.org

    Args:
        home (``Text``): Home location for the identifier.

            * ``"inspire/arxiv"``: retreive information from inspire using
                arXiv number.
            * ``"inspire/doi"``: retreive information from inspore using
                doi number.
            * ``"doi"``: retreive information from doi.org
            * ``"zenodo"``: retreive information from zenodo using zenodo
                identifier.

        identifier (``Text``): web identifier
        timeout (``int``, default ``5``): time out.

    Returns:
        ``Text``:
        Bibtex entry.
    """
    # pylint: disable=W1203

    home_loc = {
        "inspire/arxiv": "https://inspirehep.net/api/arxiv/%s",
        "inspire/doi": "https://inspirehep.net/api/doi/%s",
        "doi": "https://doi.org/%s",
        "zenodo": "https://zenodo.org/api/records/%s",
    }

    try:
        response = requests.get(
            home_loc[home] % identifier,
            headers={"accept": "application/x-bibtex"},
            timeout=timeout,
        )
        if not 200 <= response.status_code <= 299:
            log.debug(f"HTML Status Code: {response.status_code}")
            return ""
        response.encoding = "utf-8"
        return response.text
    except (requests.ConnectionError, requests.ConnectTimeout) as err:
        raise ConnectionError(
            "Can not retreive BibTeX information: No internet connection."
        ) from err
    except Exception as err:  # pylint: disable=W0718
        log.debug(str(err))
        return ""


def check_updates() -> None:
    """
    Check Spey Updates.

    .. versionadded:: 0.1.6

    Spey always checks updates when initialised. To disable this set the following

    Option if using terminal:

    .. code-block:: bash

        export SPEY_CHECKUPDATE=OFF

    Option if using python interface:

    .. code-block:: python

        import os
        os.environ["SPEY_CHECKUPDATE"]="OFF"

    """
    # pylint: disable=import-outside-toplevel, W1203
    try:
        response = requests.get("https://pypi.org/pypi/spey/json", timeout=1)
        response.encoding = "utf-8"
        pypi_info = response.json()
        pypi_version = pypi_info.get("info", {}).get("version", False)
        version = __version__
        if pypi_version:
            log.debug(f"Curernt version {version}, latest version {pypi_version}.")
            if "beta" in Version(version).prerelease:
                log.warning(
                    f"A prerelease version of Spey is in use: {version}. "
                    f"Latest stable version is {pypi_version}."
                )
            elif Version(version) < Version(pypi_version):
                log.warning(
                    f"A newer version ({pypi_version}) of Spey is available. "
                    f"Current version is {version}."
                )
            elif Version(version) > Version(pypi_version):
                log.warning(
                    f"An unstable version of Spey ({version}) is being used."
                    f" Latest stable version is {pypi_version}."
                )

    except Exception as err:  # pylint: disable=W0718
        # Can not retreive updates
        log.debug(str(err))
